"""
Layer.py
Will store the Layer class for Synfig layers
"""

import sys
import settings
from common.Param import Param
sys.path.append("..")


class Layer:
    """
    Class to keep Synfig Layers
    """
    def __init__(self, layer, parent_canvas):
        """
        """
        self.parent_canvas = parent_canvas
        self.layer = layer
        self.params = {}
        self.extract_params(self.params)
        self.set_description()


    def set_lottie_layer(self, lottie):
        """
        Stores the path to lottie format of layer
        """
        self.lottie_layer = lottie

    def get_lottie_layer(self):
        """
        Returns the path to lottie format of layer
        """
        return self.lottie_layer

    def get_layer(self):
        """
        Returns the original address of the layer
        This function might not be needed once class `Layer` is fully embedded
        in the converter
        """
        return self.layer

    def get_type(self):
        """
        Returns the type of layer
        """
        if "type" in self.layer.keys():
            return self.layer.attrib["type"]
        return settings.UNKNOWN_LAYER

    def add_param(self, key, param):
        """
        param is of type __something__
        Manually adding some parameter
        """
        self.params[key] = Param(param, self)
        return self.params[key]

    def set_group_params(self):
        """
        If the current layer is group, it sets the immediate children parameters
        as group_children=1. This is needed in convert methods like
        radial_composite and bone_link
        """
        typ = self.get_type()
        if typ in settings.GROUP_LAYER or typ in settings.PRE_COMP_LAYER:
            for param in self.params.values():
                param.is_group_child = True

    def extract_params(self, params):
        """
        Will extract the parameters from the layer and store in params
        """
        for child in self.layer:
            if child.tag == "param":
                key = child.attrib["name"]
                params[key] = Param(child, self)
        self.set_group_params()

    def get_param(self, *keys):
        """
        Given a key, this will return the address of the child corresponding to
        it
        If multiple keys are given, it will return the key found first
        """
        for key in keys:
            if key in self.params.keys():
                return self.params[key]
        return Param(None, None)
        #raise KeyError("No parameter %s in Layer %s", key, self.get_type())

    def is_active(self):
        """
        Returns true if a layer is active
        """
        if self.layer.attrib["active"] == "true":
            return True
        return False

    def to_render(self):
        """
        Returns true if we have to render the layer
        """
        key = "exclude_from_rendering"
        if key in self.layer.keys() and self.layer.attrib[key] == "true":
            return False
        return True

    def set_description(self):
        """
        Sets the layer description if given, otherwise the default description
        is set
        """
        if "desc" in self.layer.keys():
            self.description = self.layer.attrib["desc"]
        else:
            # Set the default name
            if self.get_type() in settings.SHAPE_LAYER:
                desc = settings.LAYER_SHAPE_NAME
            elif self.get_type() in set.union(settings.SOLID_LAYER, settings.SHAPE_SOLID_LAYER):
                desc = settings.LAYER_SOLID_NAME
            elif self.get_type() in settings.IMAGE_LAYER:
                desc = settings.LAYER_IMAGE_NAME
            elif self.get_type() in set.union(settings.PRE_COMP_LAYER, settings.GROUP_LAYER):
                desc = settings.LAYER_PRECOMP_NAME
            else:
                desc = settings.UNKNOWN_LAYER
            desc += str(settings.layer_count.inc())
            self.description = desc

    def get_description(self):
        """
        Getter for description of the layer
        """
        return self.description

    def getparent(self):
        """
        Returns the parent of this Layer
        return type: common.Canvas.Canvas
        """
        return self.parent_canvas

    def add_offset(self):
        """
        Inserts necassary offset in the positions of the layers if they lie inside
        another composition of Lottie
        """
        # This if condition is not applicable for group, rotate, precomp... layers
        if not settings.INSIDE_PRECOMP:
            return

        update_dict = []
        compare = {"center", "origin", "point1", "point2", "tl", "br", "p1", "p2"}
        for key in compare:
            param = self.get_param(key)
            if param.get() is not None:
                update_dict.append(param)

        for param in update_dict:
            param.add_offset()



"""
DashItem.py
To store a single Dash Item, this class will be used

REFER HERE:
https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/dashitem.cpp
"""

import sys
sys.path.append("..")


class DashItem:
    """
    Class to store Synfig format Dash Items
    """
    def __init__(self, offset, length, sidebefore, sideafter):
        self.offset_ = offset
        self.length_ = length
        self.side_type_ = []
        self.side_type_.append(sidebefore)
        self.side_type_.append(sideafter)

    def set_position(self, x):
        self.position_ = x

    def get_side_type_before(self):
        return self.side_type_[0]

    def get_side_type_after(self):
        return self.side_type_[1]

    def get_length(self):
        return self.length_

    def get_offset(self):
        return self.offset_



# pylint: disable=line-too-long
"""
Will store all the functions corresponding to Image Layer in lottie
"""

import sys
import copy
from lxml import etree
import settings
from helpers.transform import gen_helpers_transform
from helpers.blendMode import get_blend
from common.Param import Param
from sources.image import add_image_asset
from shapes.rectangle import to_Synfig_axis
sys.path.append("..")


def gen_layer_image(lottie, layer, idx):
    """
    Generates the dictionary corresponding to layers/image.json

    Args:
        lottie (dict)       : Lottie generated image stored here
        layer  (common.Layer.Layer) : Synfig format image layer
        idx    (int)        : Stores the index(number of) of image layer

    Returns:
        (None)
    """
    layer.add_offset()

    lottie["ddd"] = settings.DEFAULT_3D
    lottie["ind"] = idx
    lottie["ty"] = settings.LAYER_IMAGE_TYPE
    lottie["nm"] = layer.get_description()
    lottie["sr"] = settings.LAYER_DEFAULT_STRETCH
    lottie["ks"] = {}   # Transform properties to be filled

    settings.lottie_format["assets"].append({})
    st = add_image_asset(settings.lottie_format["assets"][-1], layer)
    asset = settings.lottie_format["assets"][-1]

    # setting class (jpg, png)
    lottie["cl"] = asset["p"].split(".")[-1]

    # setting the reference id
    lottie["refId"] = asset["id"]

    st["tl"].animate("vector")
    st["br"].animate("vector")

    st["scale"] = gen_image_scale(st["tl"], st["br"], asset["w"], asset["h"])
    # Animation of this scale is needed again, as helpers/transform does not do
    # path calculation again
    st["scale"].animate("image_scale")

    anchor = settings.DEFAULT_ANCHOR
    rotation = settings.DEFAULT_ROTATION
    opacity = layer.get_param("amount")
    opacity.animate("opacity")

    gen_helpers_transform(lottie["ks"], st["tl"], anchor, st["scale"], rotation, opacity)


    lottie["ao"] = settings.LAYER_DEFAULT_AUTO_ORIENT

    lottie["ip"] = settings.lottie_format["ip"]
    lottie["op"] = settings.lottie_format["op"]
    lottie["st"] = 0            # Don't know yet
    get_blend(lottie, layer)


def gen_image_scale(animated_1, animated_2, width, height):
    """
    In Synfig, no scale parameter is available for image layer, so it will be
    created here for Lottie conversion

    Args:
        animated_1 (common.Param.Param): point1 animation in Synfig format
        animated_2 (common.Param.Param): point2 animation in Synfig format
        width      (int)                : Width of the original image
        height     (int)                : Height of the original image

    Returns:
        (lxml.etree._Element) : Scale parameter in Synfig format
    """
    st = '<param name="image_scale"><real value="0.0000000000"/></param>'
    root = etree.fromstring(st)
    image_scale = Param(root, None)
    #image_scale.animate("image_scale")
    image_scale.animate_without_path("image_scale")


    window = {}
    window["first"] = sys.maxsize
    window["last"] = -1

    animated_1.update_frame_window(window)
    animated_2.update_frame_window(window)
    # Minimizing the window size
    if window["first"] == sys.maxsize and window["last"] == -1:
        window["first"] = window["last"] = 0
    fr = window["first"]

    # Filling the first 2 frames with there original scale values
    fill_image_scale_at_frame(image_scale[0], animated_1, animated_2, width, height, fr)
    fill_image_scale_at_frame(image_scale[0], animated_1, animated_2, width, height, fr + 1)
    fr += 2

    while fr <= window["last"]:
        new_waypoint = copy.deepcopy(root[0][0])
        time = fr / settings.lottie_format["fr"]
        time = str(time) + "s"
        new_waypoint.attrib["time"] = time
        root[0].append(new_waypoint)
        fill_image_scale_at_frame(image_scale[0], animated_1, animated_2, width, height, fr)
        fr += 1
    return image_scale


def fill_image_scale_at_frame(scale_animated, animated_1, animated_2, width, height, frame):
    """
    Generates the scale at a given frame according to point1 and point2 in
    comparison with original width and height of image

    Args:
        scale_animated (lxml.etree._Element) : Scale animation in Synfig format
        animated_1     (common.Param.Param) : point1 animation in Synfig format
        animated_2     (common.Param.Param) : point2 animation in Synfig format
        anim1_path     (dict)                : point1 animation in Lottie format
        anim2_path     (dict)                : point2 animation in Lottie format
        width          (int)                 : Width of original image
        height         (int)                 : Height of original image
        frame          (int)                 : Frame at which scale is to generated

    Returns:
        (None)
    """
    pos1 = animated_1.get_value(frame)
    pos2 = animated_2.get_value(frame)
    pos1, pos2 = to_Synfig_axis(pos1, "vector"), to_Synfig_axis(pos2, "vector")
    pos1 = [x * settings.PIX_PER_UNIT for x in pos1]
    pos2 = [x * settings.PIX_PER_UNIT for x in pos2]

    scale_x = (pos2[0] - pos1[0]) * 100 / width
    scale_y = (pos1[1] - pos2[1]) * 100 / height

    # Assumption: all frames till the maximum are present in the animation
    scale_animated[frame][0].attrib["value"] = str(scale_x)
    scale_animated[frame][0].attrib["value2"] = str(scale_y)
    scale_animated[frame].attrib["before"] = "linear"
    scale_animated[frame].attrib["after"] = "linear"



"""
Activepoint.py
To store if a point is active at a time T or not

REFER HERE:
https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/activepoint.h
"""

import sys
import settings
from canvas import convert_time_to_frames
from functools import total_ordering
sys.path.append("..")


@total_ordering
class Activepoint:
    """
    Class to store the active point of Synfig
    """
    def __init__(self, time, state, priority = 0):
        self.state = state
        self.priority = priority

        # Convert time to frames
        if time == 'SOT':
            self.time = settings.SOT
        elif time == 'EOT':
            self.time = settings.EOT
        else:
            self.time = convert_time_to_frames(time)

    def __lt__(self, rhs):
        return ((self.time) < (rhs.time))

    def __eq__(self, rhs):
        return ((self.time) == (rhs.time))

    def get_time(self):
        return self.time



# pylint: disable=line-too-long
"""
Module contains all functions required for calculation/help in bezier related
queries
"""

import sys
sys.path.append("..")


def get_bezier_time(P0, P1, P2, P3, point, frames):
    """
    Returns the fraction of time at which the value of bezier curve equals to
    the value of point
    B(t) = (1 - t)**3 P0 + 3(1 - t)**2 tP1 + 3(1 - t)t**2 P2 + t**3 P3
    where 0 < t < 1
    To optimize, binary search could be used here or later numpy might be used

    Args:
        P0 (float | common.Vector.Vector | common.Color.Color) : First control point
        P1 (float | common.Vector.Vector | common.Color.Color) : Second control point
        P2 (float | common.Vector.Vector | common.Color.Color) : Third control point
        P3 (float | common.Vector.Vector | common.Color.Color) : Fourth control point
        point (float | common.Vector.Vector | common.Color.Color) : point at which bezier curve is equal to this point
        frames (int) : Total number of frames present in this curve

    Returns:
        (float) : Time at which the point is present on the bezier curve
    """
    min_diff = sys.maxsize
    fr = 1  # On frame number 1
    ret = 0
    while fr <= frames:
        t = fr/frames
        val = get_bezier_val(P0, P1, P2, P3, t)
        if abs(val - point) < min_diff:
            min_diff = abs(val - point)
            ret = t
        fr += 1
    return ret


def get_bezier_val(P0, P1, P2, P3, t):
    """
    Returns the value of bezier function at time t

    Args:
        P0 (float | common.Vector.Vector | common.Color.Color) : First control point
        P1 (float | common.Vector.Vector | common.Color.Color) : Second control point
        P2 (float | common.Vector.Vector | common.Color.Color) : Third control point
        P3 (float | common.Vector.Vector | common.Color.Color) : Fourth control point
        t  (float)                            : Time

    Returns:
        (float | common.Vector.Vector | common.Color.Color) : value of bezier curve at time t
    """
    bezier = (((1 - t)**3) * P0) + (3*((1 - t)**2) * t*P1) + (3*(1 - t)*(t**2)*P2) + ((t**3)*P3)
    return bezier



# pylint: disable=line-too-long
"""
Vector.py
This module will store the Vector class
"""

import sys
import math
import common
sys.path.append("..")


class Vector:
    """
    To store the position of layers
    val1 represents the x-axis value
    val2 represents the y-axis value

    For other parameters
    val1 represents the value of the parameter
    val2 represents the time parameter

    type represents what this vector is representing
    """

    def __init__(self, val1=0, val2=0, _type=None):
        """
        Args:
            val1  (float) : First value of the vector
            val2  (float) : Second value of the vector
            _type (:obj: `str`, optional)  : Type of vector

        Returns:
            (None)
        """
         
        #print(type(val2), isinstance(val2, common.Angle.Angle))
        if isinstance(val2, common.Angle.Angle):
            self.val1 = val1 * common.Angle.CosAngle(val2).get()
            self.val2 = val1 * common.Angle.SinAngle(val2).get()
            self.type = _type
        else:
            self.val1 = val1
            self.val2 = val2
            self.type = _type

    def __str__(self):
        return "({0},{1}, {2})".format(self.val1, self.val2, self.type)

    def __add__(self, other):
        val1 = self.val1 + other.val1
        val2 = self.val2 + other.val2
        return Vector(val1, val2, self.type)

    def __sub__(self, other):
        val1 = self.val1 - other.val1
        val2 = self.val2 - other.val2
        return Vector(val1, val2, self.type)

    def __neg__(self):
        return -1 * self

    def __getitem__(self, key):
        if key:
            return round(self.val2, 3)
        return round(self.val1, 3)

    def __setitem__(self, key, value):
        if key:
            self.val2 = value
        else:
            self.val1 = value

    def isnan(self):
        """
        Returns true is any value of this vector is nan
        """
        return math.isnan(self.val1) or math.isnan(self.val2)

    def mag(self):
        """
        Returns the magnitude of the vector

        Args:
            (None)
        Returns:
            (float) : The magnitude of the vector
        """
        return math.sqrt(self.mag_squared())

    def inv_mag(self):
        """
        Returns the inverse of the magnitude of the vector

        Args:
            (None)
        Returns:
            (float) : Magnitude inversed
        """
        if self.mag() == 0:
            return float('nan')
        return 1.0 / self.mag()

    def perp(self):
        """
        Returns a perpendicular version of the vector

        Args:
            (None)
        Returns:
            (common.Vector.Vector) : Perpendicular vector with same magnitude
        """
        return Vector(self.val2, -self.val1)

    def is_equal_to(self, other):
        """
        Tells if the current vector is equal to `other` vector

        Args:
            other (common.Vector.Vector) : The vector to be compared with

        Returns:
            (bool) : True if the vectors are equal
                   : False otherwise
        """
        return common.misc.approximate_equal((self - other).mag_squared(), 0)

    def mag_squared(self):
        """
        Returns the squared magnitude of the vector

        Args:
            (None)
        Returns:
            (float) : squared magnitude
        """
        ret = self.val1 * self.val1 + self.val2 * self.val2
        return ret

    def angle(self):
        """
        Returns the angle of this Vector in Radians

        Args:
            (None)
        Returns:
            (common.Angle.RadAngle): The angle of Vector
        """
        return common.Angle.RadAngle(math.atan2(self.val2, self.val1))

    def norm(self):
        """
        Returns a normalised version of the vector

        Args:
            (None)
        Returns:
            (common.Vector.Vector) : itselves whose magnitude is 1
        """
        obj = self * self.inv_mag()
        self.__dict__.update(obj.__dict__)
        return self

    # other can only be of type real
    def __mul__(self, other):
        if not isinstance(other, self.__class__):
            val1 = self.val1 * other
            val2 = self.val2 * other
            return Vector(val1, val2, self.type)
        elif isinstance(other, self.__class__):
            return self.val1*other.val1 + self.val2*other.val2
        raise Exception('Multiplication with {} not defined'.format(type(other)))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if not isinstance(other, self.__class__):
            val1 = self.val1 / other
            val2 = self.val2 / other
            return Vector(val1, val2, self.type)
        raise Exception('Division with {} not defined'.format(type(other)))

    def get_list(self):
        """
        Get val1 and val2 values in the format required by lottie

        Args:
            (None)

        Returns:
            (list) : Contains the Vector in list format
        """
        return [self.val1, self.val2]

    def get_val(self):
        """
        Get value in the format required by lottie

        Args:
            (None)

        Returns:
            (list) : Depending upon _type a list is returned
        """
        if self.type == "origin":
            ret = [self.val1, self.val2]
        elif self.type == "circle_radius":
            ret = [self.val1, self.val1]
        elif self.type in {"rectangle_size", "image_scale", "scale_layer_zoom", "group_layer_scale", "stretch_layer_scale"}:
            ret = [self.val1, self.val3]
        else:
            ret = [self.val1]
        return ret

    def add_new_val(self, val3):
        """
        This function store an additional value in the vector.
        This is currently required by the rectangle layer

        Args:
            val3 (float) : Some Vectors need additional value to be used later

        Returns:
            (None)
        """
        self.val3 = val3

    def set_type(self, _type):
        """
        This set's the type of the Vector

        Args:
            _type (str) : Type of Vector to be set

        Returns:
            (None)
        """
        self.type = _type



"""
This module will store all the functions required for opacity property of lottie
"""

import sys
import settings
sys.path.append("../")


def gen_effects_opacity(lottie, layer, idx):
    """
    Generates the dictionary corresponding to effects/opacity.json

    Args:
        lottie (dict)               : Lottie format effects stored in this
        layer  (common.Layer.Layer)  : Synfig format layer
        idx    (int)                : Index/Count of effect

    Returns:
        (None)
    """
    lottie["ty"] = settings.EFFECTS_OPACITY     # Effect type
    lottie["nm"] = "Opacity"                    # Name
    lottie["ix"] = idx                          # Index
    lottie["v"] = {}                            # Value of opacity

    opacity = layer.get_param("amount")
    opacity.animate("effects_opacity")
    opacity.fill_path(lottie, "v")



# pylint: disable=line-too-long
"""
Main driver functions to generate layers
"""

import sys
import logging
import settings
from layers.shape import gen_layer_shape
from layers.solid import gen_layer_solid
from layers.image import gen_layer_image
from layers.shape_solid import gen_layer_shape_solid
from layers.preComp import gen_layer_precomp
from layers.group import gen_layer_group
from layers.blur import gen_layer_blur

sys.path.append("..")

def blur_test(lottie):
	"""
	This function will test if this layer has already been blurred or not

	Args:
		lottie (dict) : Lottie dictionary of effects in a layer

	Returns:
		(True if layer has already been blurred)
	"""

	for effects in lottie:
		if effects["ty"] == 29:
			return True

	return False

def calculate_blurs_needed(itr):
	"""
	This function will be called for each non blur layer to calculate all blur layers above it.

	Args:
		itr (int) : Position of layer in canvas

	Returns:
		(None)
	"""
	settings.non_blur_dictionary[itr] = [layer for layer in settings.blur_dictionary.keys() if layer < itr]

def append_blur_dict(layer,itr,group_flag):
	"""
	This function will generate and assign the gaussian blur dictionary required for different layers

	Args:
		layer (common.Layer.Layer) : Synfig format layer
		itr   (int)                : Position of layer in canvas
		group_flag (boolean)	   : True if layer belongs to a Group Layer

	Returns:
		(None)
	"""
	blur_dict = []
	layers = [settings.blur_dictionary[layer_var] for layer_var in settings.non_blur_dictionary[itr]]
	gen_layer_blur(blur_dict,layers)
	if len(layers)!=0:
		if group_flag:
			for asset_index,_ in enumerate(settings.lottie_format["assets"]):
				if "layers" in _.keys():
					for index,val in enumerate(settings.lottie_format["assets"][asset_index]["layers"]):
						blur_flag = blur_test(settings.lottie_format["assets"][asset_index]["layers"][index]["ef"])
						if not blur_flag:
							for blur in blur_dict:
								settings.lottie_format["assets"][asset_index]["layers"][index]["ef"].append(blur)

		else:
			for index,val in enumerate(settings.lottie_format["layers"]):
				if settings.lottie_format["layers"][index]["nm"] == layer.get_description():
					for blur in blur_dict:
						settings.lottie_format["layers"][index]["ef"].append(blur)

def gen_layers(lottie, canvas, layer_itr):
	"""
	This function will be called for each canvas/composition. Main function to
	generate all the layers

	Args:
		lottie (dict) : Layers in Lottie format
		canvas (common.Canvas.Canvas) : Synfig format canvas
		layer_itr (int) : position of layer in canvas

	Returns:
		(None)
	"""
	itr = layer_itr
	shape = settings.SHAPE_LAYER
	solid = settings.SOLID_LAYER
	shape_solid = settings.SHAPE_SOLID_LAYER
	image = settings.IMAGE_LAYER
	pre_comp = settings.PRE_COMP_LAYER
	group = settings.GROUP_LAYER
	skeleton = settings.SKELETON_LAYER
	blur = settings.BLUR_LAYER
	supported_layers = set.union(shape, solid, shape_solid, image, pre_comp, group, skeleton,blur)
	if settings.WITHOUT_VARIABLE_WIDTH:
		shape.add("outline")
		settings.WITHOUT_VARIABLE_WIDTH = False
		
	while itr >= 0:
		layer = canvas[itr]
		if layer.get_type() not in supported_layers:  # Only supported layers
			logging.warning(settings.NOT_SUPPORTED_TEXT, layer.get_type())
			itr -= 1
			continue
		elif not layer.is_active():   # Only render the active layers
			logging.info(settings.NOT_ACTIVE_TEXT, layer.get_type())
			itr -= 1
			continue
		elif not layer.to_render():   # If we don't have to render the layer
			logging.info(settings.EXCLUDE_FROM_RENDERING, layer.get_type())
			itr -= 1
			continue

		if layer.get_type() != "blur":
			lottie.append({})
			layer.set_lottie_layer(lottie[-1])

		if layer.get_type() in shape:           # Goto shape layer
			gen_layer_shape(lottie[-1],
							layer,
							itr)
			calculate_blurs_needed(settings.LEVEL)
			append_blur_dict(layer,settings.LEVEL,settings.INSIDE_PRECOMP)

		elif layer.get_type() in solid:         # Goto solid layer
			gen_layer_solid(lottie[-1],
							layer,
							itr)
			calculate_blurs_needed(settings.LEVEL)
			append_blur_dict(layer,settings.LEVEL,settings.INSIDE_PRECOMP)

		elif layer.get_type() in shape_solid:   # Goto shape_solid layer
			gen_layer_shape_solid(lottie[-1],
								  layer,
								  itr)
			calculate_blurs_needed(settings.LEVEL)
			append_blur_dict(layer,settings.LEVEL,settings.INSIDE_PRECOMP)

		elif layer.get_type() in image:   # Goto image layer
			gen_layer_image(lottie[-1],
							layer,
							itr)
			calculate_blurs_needed(settings.LEVEL)
			append_blur_dict(layer,settings.LEVEL,settings.INSIDE_PRECOMP)

		elif layer.get_type() in blur:
			settings.blur_dictionary[settings.LEVEL] = layer

		elif layer.get_type() in pre_comp:      # Goto precomp layer
			gen_layer_precomp(lottie[-1],
							  layer,
							  itr)
			return  # other layers will be generated inside the precomp
		elif layer.get_type() in group:       # Goto group layer
			gen_layer_group(lottie[-1],
							layer,
							itr)
			# No return statement here
		elif layer.get_type() in skeleton:
			pass
			# skeletons are just for linking purposes which is served by bones

		settings.LEVEL += 1
		itr -= 1



"""
Will store all the functions needed to export the scale layer
"""

import sys
import copy
import settings
from helpers.transform import gen_helpers_transform
sys.path.append("..")


def gen_layer_scale(lottie, layer, layer_name="scale_layer"):
    """
    Help generate transform properties of a scale layer

    Args:
        lottie (dict) : Store transform properties in lottie format
        layer  (common.Layer.Layer) Transform properties in Synfig format

    Returns:
        (None)
    """
    animation_type = "scale_layer_zoom"
    if layer_name == "stretch_layer":
        animation_type = "stretch_layer_scale"

    center = layer.get_param("center")
    center.animate("vector")
    anchor = copy.deepcopy(center)
    # deep copy changes address of parent layer also
    anchor.parent = center.parent
    pos = center

    scale = layer.get_param("amount")  # This is scale amount
    scale.animate(animation_type)

    anchor.add_offset()
    if settings.INSIDE_PRECOMP:
        pos.add_offset()
    anchor.animate("vector", True)
    pos.animate("vector", True)
    gen_helpers_transform(lottie, pos, anchor, scale)



# pylint: disable=line-too-long
"""
Implements all the functions required for generating a lottie mask
"""

import sys
import settings
from common.Count import Count
from properties.shapeKeyframed import gen_properties_shapeKeyframed
from properties.value import gen_properties_value
sys.path.append("../")


def gen_mask(lottie, invert, bline_point, idx):
    """
    Generates a mask. This was specifically needed to support the invert
    parameter

    Args:
        lottie      (dict) : The final mask will be stored in this dict
        invert      (bool) : Tells if the invert parameter is set or not
        bline_point (common.Layer.Layer | common.Param.Param) : Synfig format shape
        idx         (int)  : Specifies the index of this mask

    Returns:
        (None)
    """
    lottie["inv"] = invert
    lottie["mode"] = settings.MASK_ADDITIVE
    lottie["pt"] = {}
    lottie["o"] = {}    # Opacity
    lottie["x"] = {}    # Expression
    lottie["nm"] = "Mask " + str(idx)
    index = Count()

    gen_properties_shapeKeyframed(lottie["pt"], bline_point, index.inc())
    gen_properties_value(lottie["o"], settings.DEFAULT_OPACITY, index.inc(), settings.DEFAULT_ANIMATED, settings.NO_INFO)
    gen_properties_value(lottie["x"], 0, index.inc(), settings.DEFAULT_ANIMATED, settings.NO_INFO)

    convert_non_loop(lottie["pt"])


def convert_non_loop(lottie):
    """
    Converts the non loop shapes to looped shapes by changing the in and out
    tangents accordingly

    Args:
        lottie (dict) : The shape to be modified

    Returns:
        (None)
    """
    for chld in lottie["k"]:
        if "s" in chld.keys() and "e" in chld.keys():
            char_dict = {"s", "e"}
            for char in char_dict:
                posi = chld[char][0]
                if not posi["c"]:
                    posi["o"][-1][0] = posi["o"][-1][1] = 0
                    posi["i"][0][0] = posi["i"][0][1] = 0
                    posi["c"] = True



"""
This module converts the canvas to lottie format
"""
import settings
from common.misc import calculate_pixels_per_unit


def convert_time_to_frames(time):
    """
    Given time in hours, minutes, seconds, and frames, split it and convert it
    to frames

    Args:
        time (str) : Time in h/m/s/f

    Returns:
        (float)
    """
    time = time.strip().split(" ")
    ret = 0
    for frame in time:
        # Adding time in hours
        if frame[-1] == "h":
            ret += float(frame[:-1]) * 60 * 60 * settings.lottie_format["fr"]
        # Adding time in minutes
        elif frame[-1] == "m":
            ret += float(frame[:-1]) * 60 * settings.lottie_format["fr"]
        # Adding time in seconds
        elif frame[-1] == "s":
            ret += float(frame[:-1]) * settings.lottie_format["fr"]
        # Adding time in frames
        elif frame[-1] == "f":
            ret += float(frame[:-1])

    return ret


def calc_time(root, lottie, which):
    """
    Converts the starting time and ending time to lottie format

    Args:
        root   (lxml.etree._Element) : Synfig format animation file
        lottie (dict)                : Lottie format animation file
        which  (str)                 : Differentiates between in time and out time

    Returns:
        (None)
    """
    if which == "ip":
        phase = "begin-time"
    elif which == "op":
        phase = "end-time"
    time = root.attrib[phase]
    lottie[which] = convert_time_to_frames(time)

    # To support canvas with single frames
    if which == "op":
        lottie[which] += 1


def gen_canvas(lottie, root):
    """
    Generates the canvas for the lottie format
    It is the outer most dictionary in the lottie json format

    Args:
        lottie (dict)               : Lottie format animation file
        root   (lxml.etree._Element): Synfig format animation file

    Returns:
        (None)
    """
    settings.view_box_canvas["val"] = [float(itr) for itr in root.attrib["view-box"].split()]
    if "width" in root.attrib.keys():
        lottie["w"] = int(root.attrib["width"])
    else:
        lottie["w"] = settings.DEFAULT_WIDTH

    if "height" in root.attrib.keys():
        lottie["h"] = int(root.attrib["height"])
    else:
        lottie["h"] = settings.DEFAULT_HEIGHT

    settings.ADDITIONAL_PRECOMP_WIDTH = 4*lottie["w"]
    settings.ADDITIONAL_PRECOMP_HEIGHT = 4*lottie["h"]

    # gamma correction values
    if "gamma-r" in root.attrib.keys():
        settings.GAMMA[0] = float(root.attrib["gamma-r"])
    if "gamma-g" in root.attrib.keys():
        settings.GAMMA[1] = float(root.attrib["gamma-g"])
    if "gamma-b" in root.attrib.keys():
        settings.GAMMA[2] = float(root.attrib["gamma-b"])

    name = settings.DEFAULT_NAME
    for child in root:
        if child.tag == "name":
            name = child.text
            break
    lottie["nm"] = name
    lottie["ddd"] = settings.DEFAULT_3D
    lottie["v"] = settings.LOTTIE_VERSION
    lottie["fr"] = float(root.attrib["fps"])
    lottie["assets"] = []       # Creating array for storing assets
    lottie["markers"] = []      # Creating array for storing markers
    calc_time(root, lottie, "ip")
    calc_time(root, lottie, "op")
    calculate_pixels_per_unit()



# pylint: disable=line-too-long
"""
This module contains all the global variables and constants
"""

from common.Count import Count

# Constants
FLOAT_PRECISION = 3
LOTTIE_VERSION = "5.3.4"
IN_POINT = 0
OUT_POINT = 1.00000004073083
DEFAULT_WIDTH = 480
DEFAULT_HEIGHT = 270
DEFAULT_NAME = "Synfig Animation"
DEFAULT_3D = 0
DEFAULT_BLEND = 0
LAYER_SHAPE_TYPE = 4
LAYER_SHAPE_NAME = "Shape Layer "
LAYER_SOLID_TYPE = 1
LAYER_SOLID_NAME = "Solid Layer "
LAYER_IMAGE_TYPE = 2
LAYER_IMAGE_NAME = "Image Layer "
LAYER_PRECOMP_TYPE = 0
LAYER_PRECOMP_NAME = "Pre Comp Layer "
CANVAS_NAME = "Canvas "
LAYER_DEFAULT_STRETCH = 1
LAYER_DEFAULT_AUTO_ORIENT = 0
OPACITY_CONSTANT = 100
DEFAULT_ANIMATED = 0
NO_INFO = "no_info"
DEFAULT_ROTATION = 0
DEFAULT_OPACITY = 100
DEFAULT_DIRECTION = 1
DEFAULT_POSITION = [0, 0]
DEFAULT_ANCHOR = [0, 0, 0]
DEFAULT_SCALE = [100, 100, 100]
DEFAULT_SKEW = 0
GAMMA = [2.2, 2.2, 2.2]     # Default RGB gamma correction values
PIX_PER_UNIT = 0
TANGENT_FACTOR = 3.0
IN_TANGENT_X = 0.58
IN_TANGENT_Y = 1
OUT_TANGENT_X = 0.42
OUT_TANGENT_Y = 0
EFFECTS_FILL = 21
EFFECTS_FILL_MASK = 10
EFFECTS_ALL_MASK = 7
EFFECTS_COLOR = 2
EFFECTS_INVERT = 7  # same as All mask, don't know why
EFFECTS_HFEATHER = 0    # horizontal feather
EFFECTS_VFEATHER = 0    # vertical feather
EFFECTS_OPACITY = 0     # Opacity ty = 0
EFFECTS_CONTROLLER = 5
EFFECTS_SLIDER = 0
EFFECTS_POINT = 3
MASK_ADDITIVE = "a"
INSIDE_PRECOMP = False  # specifies if we are inside a precomp or not
ADDITIONAL_PRECOMP_WIDTH = 0
ADDITIONAL_PRECOMP_HEIGHT = 0
NOT_SUPPORTED_TEXT = "Layer '%s' is not supported yet. For more information, contact us on Synfig forums or Github page"
NOT_ACTIVE_TEXT = "Layer '%s' is not active"
EXCLUDE_FROM_RENDERING = "Layer '%s' is excluded from rendering"
SHAPE_LAYER = {"simple_circle", "linear_gradient", "radial_gradient"}
BLUR_LAYER = {"blur"}
SOLID_LAYER = {"solid_color"}
SHAPE_SOLID_LAYER = {"region", "polygon", "advanced_outline", "outline", "circle", "rectangle", "filled_rectangle", "star"} 
IMAGE_LAYER = {"import"}
PRE_COMP_LAYER = {"rotate", "zoom", "translate", "stretch"}
GROUP_LAYER = {"group", "switch"}
SKELETON_LAYER = {"skeleton"}
UNKNOWN_LAYER = "unknown_layer"
CONVERT_METHODS = {"add", "atan2","average", "bone", "bone_link", "bone_root", "composite", "cos", "dotproduct", "exp", "fromint", "linear", "logarithm", "power", "radial_composite", "range", "reciprocal", "scale", "sine", "subtract", "switch", "vectorangle", "vectorlength", "vectorx", "vectory", "weighted_average"}
BONES = {"bone", "bone_root"}
DOT_FLAG = 0 #Used for the two types of dot product -> angle and real
BLUR_TYPE = 29
RANGE_FLAG = 0 #Used for if-else expressions
# Some waypoint animated definitions
ANIMATED = 2
SINGLE_WAYPOINT = 1
NOT_ANIMATED = 0
LEVEL = 0 #Indicates the depth of a layer
OUTLINE_FLAG = False #Flag to check for outline as outline needs the newer version of bodymovin.js
WAYPOINTS_LIST = []
WITHOUT_VARIABLE_WIDTH = False
ROOT_CANVAS = None

# Time.h
SOT = -32767*512
EOT = 32767*512

def init():
    """
    Initialises the final dictionary corresponding to conversion and
    also the canvas dictionary needed in misc functions

    Args:
        (None)

    Returns:
        (None)
    """
    # Final converted dictionary
    global lottie_format
    lottie_format = {}
    global view_box_canvas
    view_box_canvas = {}
    global num_images
    num_images = Count()
    global file_name
    file_name = {}
    global num_precomp
    num_precomp = Count()
    global OUTLINE_GROW    # outline grow param of group layer
    OUTLINE_GROW = [0]
    global layer_count  # will only count the layers which do not have there desc set
    layer_count = Count()
    global canvas_count # will only count the canvas which do not have any names
    canvas_count = Count()
    global controller_count # counts the slider and point effects controller
    controller_count = Count()
    global blur_dictionary #used to make a dictionary of blur layers
    blur_dictionary = {}
    global non_blur_dictionary #used to make a dictionary of all non blur layers
    non_blur_dictionary = {}



"""
DashItemList.py
Will store the Parameters class for Dash item list
"""

import sys
import common
from common.DashItem import DashItem
from common.ActivepointList import ActivepointList
from synfig.animation import to_Synfig_axis
sys.path.append("..")


class DashItemList:
    """
    Class to store Synfig format Dash items list
    """
    def __init__(self, dash_item, parent):
        """
        This parent should be parameter
        """
        self.parent = parent
        self.dash_item = dash_item
        self.type = dash_item.attrib["type"]
        self.entry_list = []
        self.extract_entries(self.entry_list)

    def get(self):
        """
        Returns the original param
        """
        return self.dash_item

    def __getitem__(self, itr):
        """
        Returns the child corresponding to itr
        """
        return self.entry_list[itr]

    def __setitem__(self, itr, val):
        """
        Sets the value of child corresponding to itr
        """
        self.entry_list[itr] = val

    def get_layer_type(self):
        """
        Recursively go till the top until layer is not reached
        """
        if isinstance(self.parent, common.Layer.Layer):
            return self.parent.get_type()
        return self.parent.get_layer_type()

    def get_layer(self):
        """
        Recursively find the layer of this bline
        """
        if isinstance(self.parent, common.Layer.Layer):
            return self.parent
        return self.parent

    def get_type(self):
        """
        Returns the type of bline
        """
        return self.type

    def get_loop(self):
        """
        Returns whether the bline is looped or not
        """
        if "loop" in self.dash_item.keys():
            v = self.dash_item.attrib["loop"]
            return v == "true"
        return False

    def extract_entries(self, entry_list):
        """
        Stores the entries in a list
        """
        for entry in self.dash_item:
            # Assuming it's child is always composite for now
            entry_list.append({})
            entry_list[-1]["ActivepointList"] = ActivepointList(entry.attrib.get("on"), entry.attrib.get("off"))

            element = entry[0]
            for param in element:
                tag = param.tag
                if tag == "animated":   # For dynamic list
                    tag = param.attrib["type"]
                entry_list[-1][tag] = common.Param.Param(param, element)

            assert(entry[0].tag == "composite")
            entry_list[-1]["composite"] = element

    def get_entry_list(self):
        """
        Returns the entry list
        """
        return self.entry_list

    def get_len(self):
        """
        Returns the number of entries
        """
        return len(self.entry_list)

    def set_entry(self, itr, entry, tag="vector"):
        """
        Sets the entry at itr to val
        """
        for param in entry:
            self.entry_list[itr][tag] = common.Param.Param(param, entry)

    def get_list_at_frame(self, fr):
        """
        Returns list of Dashitems at a particular frame
        Refer:  https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/valuenodes/valuenode_dilist.cpp#L129
        """
        dilist = []
        rising = [False]

        for entry in self.get_entry_list():
            amount = entry["ActivepointList"].amount_at_time(fr, rising)
            assert(amount >= 0)
            assert(amount <= 1)

            offset = to_Synfig_axis(entry["offset"].get_value(fr), "real")
            length = to_Synfig_axis(entry["length"].get_value(fr), "real")
            side_before = entry["side_before"].get()
            side_before = int(side_before[0].attrib["value"])
            side_after = entry["side_after"].get()
            side_after = int(side_after[0].attrib["value"])
            curr = DashItem(offset, length, side_before, side_after)

            if amount > 1 - 0.0000001:  # lgtm [py/redundant-comparison]
                dilist.append(curr)

        return dilist



"""
Canvas.py
Will store the Canvas class required for Synfig canvases
"""

import sys
import settings
from common.Layer import Layer
from common.Param import Param
sys.path.append("..")


class Canvas:
    """
    Class to keep Canvas for Synfig
    """
    def __init__(self, canvas, is_root_canvas = False):
        """
        """
        if is_root_canvas:
            settings.ROOT_CANVAS = self

        if isinstance(canvas, Param):
            self.canvas = canvas.get()[0]
            # If canvas is derived from a parameter store that parameter also
            self.parent_param = canvas
        else:
            self.canvas = canvas
        self.defs = {}
        self.extract_defs(self.defs)

        self.layers = []
        self.extract_layers(self.layers)
        self.bones = {}
        self.extract_bones(self.bones)
        self.set_name()

    def set_name(self):
        """
        Sets the name of the canvas
        """
        found = False
        for child in self.canvas:
            if child.tag == "name":
                self.name = child.text
                found = True
        if not found:
            self.name = settings.CANVAS_NAME + str(settings.canvas_count.inc())

    def get_canvas(self):
        """
        This won't be required once the canvas class is fully embedded in the
        code
        Returns the address of original canvas
        """
        return self.canvas

    def getparent_param(self):
        """
        This will return the parent parameter of the canvas
        Only when a canvas is present in a group or switch group layer
        """
        return self.parent_param

    def extract_defs(self, defs_dict):
        """
        Will store all the defs with there id's as there keys
        """
        for child in self.canvas:
            if child.tag == "defs":
                defs = child

                for child in defs:
                    key = child.attrib["id"]
                    defs_dict[key] = child

    def get_def(self, key):
        """
        Given a key, returns the corresponding child defined in the canvas
        """
        if key in self.defs.keys():
            return self.defs[key]
        return None
        #raise KeyError("No defs %s in Canvas %s", key, self.name)

    def extract_layers(self, layer_list):
        """
        Will store all the layers in a list
        """
        for child in self.canvas:
            if child.tag == "layer":
                layer_list.append(Layer(child, self))

    def extract_bones(self, bones_dict):
        """
        Will store all the bones in a dictionary
        """
        for chld in self.canvas:
            if chld.tag == "bones":
                bones = chld

                for child in bones:
                    key = child.attrib["guid"]
                    bones_dict[key] = Param(child, Param(bones, self))

    def get_bone(self, key):
        """
        Given a guid, returns the corresponding bone from the canvas
        """
        if key in self.bones.keys():
            return self.bones[key]
        return None
        #raise KeyError("No bone guid: %s in Canvas %s", key, self.name)

    def __getitem__(self, itr):
        """
        Returns the layer at itr
        """
        return self.layers[itr]

    def get_num_layers(self):
        """
        Returns the number of layers in the canvas
        """
        return len(self.layers)

    def get_layer_list(self):
        """
        Returns the list of layers in the canvas
        """
        return self.layers



"""
This module will store the expression controllers of lottie/AE
"""

import sys
import settings
sys.path.append("../")


def gen_effects_slider(lottie, value, idx):
    """
    Generates the dictionary corresponding to effects/slider.json
    """
    lottie["ty"] = settings.EFFECTS_SLIDER      # Effect type
    lottie["nm"] = "Slider" + str(idx)
    lottie["ix"] = idx
    lottie["v"] = value



"""
Module contains all functions required for setting up the blend mode of any
layer
"""

import sys
import settings
sys.path.append("..")


def get_blend(lottie, layer):
    """
    blend_map stores the mapping:
    composite  :  0,
    difference : 18,
    multiply   :  6,
    Hard light : 17,
    Luminance  : 11,
    Saturation : 10,
    Hue        :  9,
    Color      :  8,
    Darken     :  3,
    Brighten   :  2,
    Overlay    : 20,    Does not work perfectly :(
    Screen     : 16

    Args:
        lottie (dict)               : Lottie format layer
        layer  (common.Layer.Layer)  : Synfig format layer

    Returns:
        (None)
    """
    blend_map = {0 : 0, 18 : 10, 6 : 1, 17 : 8, 11 : 15, 10 : 13, 9 : 12, 8 : 14,
                 3 : 4, 2 : 5, 20 : 3, 16 : 2}

    blend = layer.get_param("blend_method").get()
    key = int(blend[0].attrib["value"])
    if key in blend_map.keys():
        lottie["bm"] = blend_map[key]
    else:
        lottie["bm"] = settings.DEFAULT_BLEND



"""
Will have all the functions required for generate the fill(color) in lottie
"""

import sys
import settings
from properties.value import gen_properties_value
from properties.valueKeyframed import gen_value_Keyframed
from common.misc import is_animated
from common.Count import Count
sys.path.append("..")


def gen_shapes_fill(lottie, layer):
    """
    Generates the dictionary corresponding to shapes/fill.json

    Args:
        lottie (dict)               : The lottie generated fill layer will be stored in it
        layer  (common.Layer.Layer)  : Synfig format fill (can be shape/solid anything, we
                                    only need color and opacity part from it) layer

    Returns:
        (None)
    """
    index = Count()
    lottie["ty"] = "fl"     # Type if fill
    lottie["c"] = {}       # Color
    lottie["o"] = {}       # Opacity of the fill layer
    
    # Color
    color = layer.get_param("color").get()
    is_animate = is_animated(color[0])
    if is_animate == settings.ANIMATED:
        gen_value_Keyframed(lottie["c"], color[0], index.inc())

    else:
        if is_animate == settings.NOT_ANIMATED:
            val = color[0]
        else:
            val = color[0][0][0]
        red = float(val[0].text)
        green = float(val[1].text)
        blue = float(val[2].text)
        red, green, blue = red ** (1/settings.GAMMA[0]), green ** (1/settings.GAMMA[1]), blue ** (1/ settings.GAMMA[2])
        alpha = float(val[3].text)
        gen_properties_value(lottie["c"],
                             [red, green, blue, alpha],
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)

    # Color Opacity
    opacity = layer.get_param("amount").get()
    is_animate = is_animated(opacity[0])
    if is_animate == settings.ANIMATED:
        # Telling the function that this is for opacity
        opacity[0].attrib['type'] = 'opacity'
        gen_value_Keyframed(lottie["o"], opacity[0], index.inc())

    else:
        if is_animate == settings.NOT_ANIMATED:
            val = float(opacity[0].attrib["value"]) * settings.OPACITY_CONSTANT
        else:
            val = float(opacity[0][0][0].attrib["value"]) * settings.OPACITY_CONSTANT
        gen_properties_value(lottie["o"],
                             val,
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)



# pylint: disable=line-too-long
"""
Store all functions corresponding to group layer in Synfig
"""

import sys
import math
import settings
from common.Param import Param
from common.Canvas import Canvas
from common.Count import Count
from common.misc import get_frame, approximate_equal, get_time
from sources.precomp import add_precomp_asset
from helpers.transform import gen_helpers_transform
from helpers.blendMode import get_blend
from synfig.animation import insert_waypoint_at_frame, to_Synfig_axis
sys.path.append("..")


def gen_layer_group(lottie, layer, idx):
    """
    Will generate a pre composition but has small differences than pre-comp layer used in
    layers/preComp.py
    This function will be used for group layer as well as switch group layer

    Args:
        lottie (dict)               : Lottie format layer will be stored here
        layer (common.Layer.Layer) : Synfig format group/switch layer
        idx   (int)                 : Index of the layer

    Returns:
        (None)
    """
    lottie["ddd"] = settings.DEFAULT_3D
    lottie["ind"] = idx
    lottie["ty"] = settings.LAYER_PRECOMP_TYPE
    lottie["sr"] = settings.LAYER_DEFAULT_STRETCH
    lottie["ks"] = {}   # Transform properties to be filled
    lottie["nm"] = layer.get_description()
    index = Count()

    # Extract parameters
    canvas = Canvas(layer.get_param("canvas"))
    origin = layer.get_param("origin")
    opacity = layer.get_param("amount")
    outline_grow = layer.get_param("outline_grow")
    time_offset = layer.get_param("time_offset")
    time_dilation = layer.get_param("time_dilation")
    transformation = layer.get_param("transformation")
    transform = transformation[0]
    try_par = Param(transform, Param(transformation.get(), layer))
    for child in transform:
        if child.tag == "scale":
            scale = Param(child, try_par)
        elif child.tag == "offset":
            pos = Param(child, try_par)
        elif child.tag == "angle":
            angle = Param(child, try_par)
        elif child.tag == "skew_angle":
            skew = Param(child, try_par)

    outline_grow.animate("real")

    origin.animate("vector")
    anchor = origin
    anchor.add_offset()

    angle.animate("rotate_layer_angle")

    pos.animate("vector")
    if settings.INSIDE_PRECOMP:
        pos.add_offset()

    scale.animate("group_layer_scale")
    # Generating animation for skew
    skew.animate("rotate_layer_angle")
    # Animating opacity
    opacity.animate("opacity")

    # Reset the animations after adding offset
    anchor.animate("vector", True)
    pos.animate("vector", True)
    # Generate the transform properties here
    gen_helpers_transform(lottie["ks"], pos, anchor, scale, angle, opacity, skew)

    # Store previous states, to be recovered at the end of group layer
    prev_state = settings.INSIDE_PRECOMP
    settings.OUTLINE_GROW.append(outline_grow)    # Storing the outline grow in settings, will be used inside child outlines

    settings.INSIDE_PRECOMP = True

    settings.lottie_format["assets"].append({})
    asset = add_precomp_asset(settings.lottie_format["assets"][-1], canvas, canvas.get_num_layers())
    lottie["refId"] = asset

    lottie["w"] = settings.lottie_format["w"] + settings.ADDITIONAL_PRECOMP_WIDTH # Experimental increase in width and height of precomposition
    lottie["h"] = settings.lottie_format["h"] + settings.ADDITIONAL_PRECOMP_HEIGHT
    lottie["ao"] = settings.LAYER_DEFAULT_AUTO_ORIENT
    lottie["ip"] = settings.lottie_format["ip"]
    lottie["op"] = settings.lottie_format["op"]
    lottie["st"] = 0            # Don't know yet
    get_blend(lottie, layer)

    # Time offset and speed
    lottie["tm"] = {}
    gen_time_remap(lottie["tm"], time_offset, time_dilation, index.inc())

    # Change opacity of layers for switch-group layers
    if layer.get_type() == "switch":
        change_opacity_switch(layer, lottie)
    # Change opacity of layers for group layers
    elif layer.get_type() == "group":
        change_opacity_group(layer, lottie)

    # Return to previous state, when we go outside the group layer
    settings.INSIDE_PRECOMP = prev_state
    settings.OUTLINE_GROW.pop()


def change_opacity_group(layer, lottie):
    """
    Will make the opacity of underlying layers 0 according to the layers lying
    inside z range(if it is active)[z-range is non-animatable]

    Args:
        layer (common.Layer.Layer) : Synfig format layer
        lottie (dict)      : Lottie format layer

    Returns:
        (None)
    """
    z_range = layer.get_param("z_range")
    z_range_pos = layer.get_param("z_range_position")
    z_range_depth = layer.get_param("z_range_depth")
    canvas = Canvas(layer.get_param("canvas"))

    for assets in settings.lottie_format["assets"]:
        if assets["id"] == lottie["refId"]:
            root = assets
            break

    # If z-range is non-active (static value)
    if z_range[0].attrib["value"] == "false":
        return

    z_range_pos.animate("real")

    z_range_depth.animate("real")

    z_st, z_en = float('-inf'), float('-inf')
    active_range = [] # Stores the time and change of layers in z-range
    fr = settings.lottie_format["ip"]
    while fr <= settings.lottie_format["op"]:
        pos_val = to_Synfig_axis(z_range_pos.get_value(fr), "real")
        depth_val = to_Synfig_axis(z_range_depth.get_value(fr), "real")
        st, en = math.ceil(pos_val), math.floor(pos_val + depth_val)
        if st > en or en < 0:
            if (fr == settings.lottie_format["ip"]) or (z_st != -1 and z_en != -1):
                z_st, z_en = -1, -1
                active_range.append([fr, z_st, z_en])
        elif (st != z_st) or (en != z_en):
            z_st, z_en = st, en
            active_range.append([fr, z_st, z_en])
        fr += 1

    z_value = 0
    for c_layer in reversed(canvas.get_layer_list()):
        if not c_layer.is_active() or not c_layer.to_render():
            continue
        active_time = set()
        itr = 0
        while itr < len(active_range):
            st, en = active_range[itr][1], active_range[itr][2]
            if z_value <= en and z_value >= st:
                now = active_range[itr][0] / settings.lottie_format["fr"]
                later = get_time_bound("op")
                if itr + 1 < len(active_range):
                    later = active_range[itr + 1][0] / settings.lottie_format["fr"]
                active_time.add((now, later))
            itr += 1
        active_time = sorted(active_time)
        inactive_time = sorted(flip_time(active_time))

        if c_layer.get_type() in set.union(settings.SHAPE_SOLID_LAYER, settings.SOLID_LAYER):
            anim_type = "effects_opacity"
            sw = 1
        elif c_layer.get_type() in set.union(settings.PRE_COMP_LAYER, settings.GROUP_LAYER, settings.IMAGE_LAYER):
            anim_type = "opacity"
            sw = 2
        elif c_layer.get_type() in settings.SHAPE_LAYER:
            anim_type = "opacity"
            sw = 3
            dic = root["layers"][z_value]["shapes"][1]["o"]

        animation = gen_hold_waypoints(inactive_time, c_layer, anim_type)
        animation.animate(anim_type)

        if sw == 1:
            animation.fill_path(root["layers"][z_value]["ef"][0]["ef"][-1], "v")
            # See effects/fill.py: Opacity is the last property, and hence we are using [-1].
            # We should actually search for "opacity", but due to multiple elements with same
            # "ty"(which should not happen), we are using [-1]. "ty" here means type which uniquely
            # identifies the effect in Lottie, but 'horizontal feather', 'vertical feather' and
            # 'opacity' in Lottie have the same type and hence we can not search for "opacity"
            # uniquely
        elif sw == 2:
            animation.fill_path(root["layers"][z_value]["ks"], "o")
        elif sw == 3:
            animation.fill_path(root["layers"][z_value]["shapes"][1], "o")

        z_value += 1



def change_opacity_switch(layer, lottie):
    """
    Will make the opacity of underlying layers 0 according to the active layer

    Args:
        layer (common.Layer.Layer) : Synfig format layer
        lottie (dict)      : Lottie format layer

    Returns:
        (None)
    """
    layer_name = layer.get_param("layer_name")
    canvas = Canvas(layer.get_param("canvas"))

    layer_name.animate_without_path("string")
    for assets in settings.lottie_format["assets"]:
        if assets["id"] == lottie["refId"]:
            root = assets
            break

    it = 0
    for c_layer in reversed(canvas.get_layer_list()):
        if not c_layer.is_active() or not c_layer.to_render():
            continue
        active_time = set()
        description = root["layers"][it]["nm"]

        waypoint_itr = 0
        while waypoint_itr < len(layer_name[0]):
            waypoint = layer_name[0][waypoint_itr]
            l_name = waypoint[0].text
            if (l_name == description) or (l_name is None and it == 0):
                update_time(active_time, layer_name[0], waypoint_itr)
            waypoint_itr += 1

        active_time = sorted(active_time)
        inactive_time = sorted(flip_time(active_time))

        sw = 0  # To decide which if condition to go to
        if c_layer.get_type() in set.union(settings.SHAPE_SOLID_LAYER, settings.SOLID_LAYER):
            anim_type = "effects_opacity"
            sw = 1
        elif c_layer.get_type() in set.union(settings.PRE_COMP_LAYER, settings.GROUP_LAYER, settings.IMAGE_LAYER):
            anim_type = "opacity"
            sw = 2
        elif c_layer.get_type() in settings.SHAPE_LAYER:
            anim_type = "opacity"
            sw = 3

        animation = gen_hold_waypoints(inactive_time, c_layer, anim_type)
        animation.animate(anim_type)

        if sw == 1:
            animation.fill_path(root["layers"][it]["ef"][0]["ef"][-1], "v")
            # See effects/fill.py: Opacity is the last property, and hence we are using [-1].
            # We should actually search for "opacity", but due to multiple elements with same
            # "ty"(which should not happen), we are using [-1]. "ty" here means type which uniquely
            # identifies the effect in Lottie, but 'horizontal feather', 'vertical feather' and
            # 'opacity' in Lottie have the same type and hence we can not search for "opacity"
            # uniquely
        elif sw == 2:
            animation.fill_path(root["layers"][it]["ks"], "o")
        elif sw == 3:
            animation.fill_path(root["layers"][it]["shapes"][1], "o")

        it += 1


def flip_time(time):
    """
    Time will be in a set();
    Example: input: ((2, 3), (4, 5))
             output: ((0, 2), (3, 4), (5, frame_last_time))

    Args:
        time (set) : Range of time is stored in this

    Returns:
        (set) : Flipped/opposite of `time` is returned
    """
    ret = set()
    last = settings.lottie_format["op"]/settings.lottie_format["fr"]
    z = 0
    for it in time:
        if (not approximate_equal(z, it[0])) and (not approximate_equal(it[0], it[1])):
            ret.add((z, it[0]))
        z = it[1]
    if not approximate_equal(z, last):
        ret.add((z, last))
    return ret


def gen_hold_waypoints(inactive_time, layer, anim_type):
    """
    Will only be used to modify opacity waypoints, and set zero values where the
    layer is inactive

    Args:
        inactive_time (set) : Range of time when the layer will be inactive
        layer (common.Layer.Layer) : Synfig format layer
        anim_type (str) : Specifies whether it is effects_opacity or opacity (it
                          will effect a factor of 100)

    Returns:
        (common.Param.Param) : Modified opacity animation is returned
    """
    opacity = layer.get_param("amount")

    opacity.animate(anim_type)
    opacity_dict = {}
    opacity_dict["o"] = {}
    opacity.fill_path(opacity_dict, "o")
    opacity_dict = opacity_dict["o"]

    for it in inactive_time:
        # First add waypoints at both points, make it constant interval
        # then remove any in-between waypoints
        first = round(it[0] * settings.lottie_format["fr"])
        second = round(it[1] * settings.lottie_format["fr"])
        insert_waypoint_at_frame(opacity[0], opacity_dict, first, anim_type)
        insert_waypoint_at_frame(opacity[0], opacity_dict, second, anim_type)

        # Making it a constant interval
        for waypoint in opacity[0]:
            if approximate_equal(get_frame(waypoint), first):
                st_waypoint = waypoint
                break
        st_waypoint.attrib["after"] = "constant"
        st_waypoint[0].attrib["value"] = str(0)

        # removing the in between waypoints
        for waypoint in opacity[0]:
            this_frame = get_frame(waypoint)
            if (not approximate_equal(this_frame, first)) and \
               (not approximate_equal(this_frame, second)) and \
               (this_frame > first and this_frame < second):
                waypoint.getparent().remove(waypoint)

    return opacity


def update_time(active_time, animated, itr):
    """
    Depending on the waypoints, the active time set is updated accordingly

    Args:
        active_time (set) : Stores the active time ranges of the layer till now
        animated (lxml.etree._Element) : Animation in Synfig format
        itr (int) : Position of the waypoint in the animation

    Returns:
        (None)
    """
    # Tuples will be added to the set
    first = get_time(animated[itr])
    if itr == 0:
        first = 0   # can use settings.lf["ip"]
    second = settings.lottie_format["op"]/settings.lottie_format["fr"]
    if itr + 1 <= len(animated) - 1:
        second = get_time(animated[itr+1])
    active_time.add((first, second))


def gen_time_remap(lottie, time_offset, time_dilation, idx):
    """
    Time offset will be converted to time remap here
    Currently time remapping will be done for each frame, but this function can
    be intelligently written only for some particular frames,hence reducing the
    space

    Args:
        lottie (dict) : Time remapping in Lottie format
        time_offset (common.Param.Param) : Offset for time in Synfig format
        time_dilation (common.Param.Param) : Speed/dilation for time in Synfig format
        idx (itr) : Index of this property in the layer

    Returns:
        (None)
    """
    time_offset.animate("time")

    time_dilation.animate("real")

    fr, lst = settings.lottie_format["ip"], settings.lottie_format["op"]
    lottie["a"] = 1 # Animated
    lottie["ix"] = idx
    lottie["k"] = []

    while fr <= lst:
        lottie["k"].append({})
        gen_dict(lottie["k"][-1], time_offset, time_dilation, fr)
        fr += 1


def gen_dict(lottie, time_offset, time_dilation, fr):
    """
    Generates the constant values for each frame

    Args:
        lottie (dict) : Bezier values will be stored in here
        offset_dict (dict) : Animation of offset in lottie format
        dilation_dict (dict) : Animation of dilation/speed in lottie format
        fr (int) : frame number

    Returns:
        (None)
    """
    lottie["i"], lottie["o"] = {}, {}
    lottie["i"]["x"], lottie["i"]["y"] = [], []
    lottie["o"]["x"], lottie["o"]["y"] = [], []
    lottie["i"]["x"].append(0.5)
    lottie["i"]["y"].append(0.5)
    lottie["o"]["x"].append(0.5)
    lottie["o"]["y"].append(0.5)
    lottie["t"] = fr

    speed_f = to_Synfig_axis(time_dilation.get_value(fr), "real")
    speed_s = to_Synfig_axis(time_dilation.get_value(fr+1), "real")
    first = time_offset.get_value(fr) + (fr/settings.lottie_format["fr"])*speed_f
    second = time_offset.get_value(fr + 1) + ((fr + 1)/settings.lottie_format["fr"])*speed_s
    first = min(max(get_time_bound("ip"), first), get_time_bound("op"))
    second = min(max(get_time_bound("ip"), second), get_time_bound("op"))

    lottie["s"], lottie["e"] = [first], [second]


def get_time_bound(st):
    """
    Returns the extreme values of time in seconds

    Args:
        st (str) : Specifies in-time or out-time

    Returns:
        (float) : Extreme value of time in seconds
    """
    ret = settings.lottie_format[st]
    if st == "op":
        ret -= 1
    ret /= settings.lottie_format["fr"]
    return ret



"""
Will store all the functions needed to export the rotate layer
"""

import sys
import copy
import settings
from helpers.transform import gen_helpers_transform
sys.path.append("..")


def gen_layer_rotate(lottie, layer):
    """
    Help generate transform properties of a rotate layer

    Args:
        lottie (dict) : Will store the transform properties in lottie format
        layer  (common.Layer.Layer) : Transform properties in Synfig format

    Returns:
        (None)
    """
    scale = settings.DEFAULT_SCALE
    origin = layer.get_param("origin")
    origin.animate("vector")
    anchor = copy.deepcopy(origin)
    # deep copy changes the parent layer also
    anchor.parent = origin.parent

    pos = origin

    amount = layer.get_param("amount")  # This is rotation amount
    amount.animate("rotate_layer_angle")

    anchor.add_offset()
    if settings.INSIDE_PRECOMP:
        pos.add_offset()
    anchor.animate("vector", True)
    pos.animate("vector", True)
    gen_helpers_transform(lottie, pos, anchor, scale, amount)



# pylint: disable=line-too-long
"""
Will store all the functions required for generation of shapeKeyframed file in
lottie
"""

import sys
from properties.shapePropKeyframe.advanced_outline import gen_bline_advanced_outline
from properties.shapePropKeyframe.outline import gen_bline_outline
from properties.shapePropKeyframe.region import gen_bline_region
from properties.shapePropKeyframe.polygon import gen_dynamic_list_polygon
from properties.shapePropKeyframe.circle import gen_list_circle
from properties.shapePropKeyframe.rectangle import gen_list_rectangle
from properties.shapePropKeyframe.star import gen_list_star
from common.Param import Param
from common.Layer import Layer
sys.path.append("../")


def gen_properties_shapeKeyframed(lottie, node, idx):
    """
    Will convert bline points/dynamic_list to bezier points as required by lottie if they are
    animated

    Args:
        lottie (dict) : Lottie generated shape keyframes will be stored here
        node   (common.Layer.Layer | common.Param.Param) : Shape/path in Synfig format :- Could be bline_point or dynamic_list
        idx    (int) : Index/Count of shape/path

    Returns:
        (None)
    """
    lottie["ix"] = idx
    lottie["a"] = 1
    lottie["k"] = []
    if isinstance(node, Layer) and node.get_type() == "circle":
        gen_list_circle(lottie["k"], node)
    elif isinstance(node, Layer) and node.get_type() in {"rectangle", "filled_rectangle"}:
        gen_list_rectangle(lottie["k"], node)
    elif isinstance(node, Layer) and node.get_type() == "star":
        gen_list_star(lottie["k"], node)
    elif isinstance(node, Layer) and node.get_type() in {"linear_gradient", "radial_gradient"}:  # rectangle layer is needed for gradients
        gen_list_rectangle(lottie["k"], node)
    elif isinstance(node, Param) and node.get_layer_type() == "region":
        gen_bline_region(lottie["k"], node)
    elif isinstance(node, Param) and node.get_layer_type() == "polygon":
        gen_dynamic_list_polygon(lottie["k"], node)
    elif isinstance(node, Param) and node.get_layer_type() == "outline":
        gen_bline_outline(lottie["k"], node)
    elif isinstance(node, Param) and node.get_layer_type() == "advanced_outline":
        gen_bline_advanced_outline(lottie["k"], node)



# pylint: disable=line-too-long
"""
Will store all the functions related to modify and play with rectangle
layer in the Synfig format
"""

import sys
import settings
from lxml import etree
sys.path.append("..")


def gen_dummy_rectangle(layer):
    """
    Generates point1 and point2 required for rectangle generation in synfig format

    Args:
        layer (common.Layer.Layer) : Synfig format layer (for now this can only be linear gradient layer)

    Returns:
        (None)
    """
    # Generate point1 and point2 to fill the whole canvas
    st = "<param name='point1'><vector><x>{x}</x><y>{y}</y></vector></param>"
    st = st.format(x=10*settings.view_box_canvas['val'][0], y=10*settings.view_box_canvas['val'][1])
    point1 = etree.fromstring(st)
    layer.add_param("point1", point1)

    st = "<param name='point2'><vector><x>{x}</x><y>{y}</y></vector></param>"
    st = st.format(x=10*settings.view_box_canvas['val'][2], y=10*settings.view_box_canvas['val'][3])
    point2 = etree.fromstring(st)
    layer.add_param("point2", point2)



"""
Will store all the functions corresponding to shapes which will be masked as
shapes in lottie
"""

import sys
import settings
from helpers.transform import gen_helpers_transform
from helpers.blendMode import get_blend
from helpers.mask import gen_mask
from common.misc import get_color_hex, is_animated
from common.Count import Count
from effects.fill import gen_effects_fill
from synfig.group import get_additional_width, get_additional_height
sys.path.append("..")


def gen_layer_shape_solid(lottie, layer, idx):
    """
    Generates the dictionary corresponding to layers/shapes.json

    Args:
        lottie (dict)       : Lottie generated solid layer stored here
        layer  (common.Layer.Layer) : Synfig format solid layer
        idx    (int)        : Stores the index(number of) of solid layer

    Returns:
        (None)
    """
    layer.add_offset()

    # Setting the solid layer which will be masked
    index = Count()
    lottie["ddd"] = settings.DEFAULT_3D
    lottie["ind"] = idx
    lottie["ty"] = settings.LAYER_SOLID_TYPE
    lottie["nm"] = layer.get_description()
    lottie["sr"] = settings.LAYER_DEFAULT_STRETCH
    lottie["ks"] = {}   # Transform properties to be filled
    lottie["ef"] = []   # Stores the effects

    pos = [settings.lottie_format["w"]/2 + get_additional_width()/2,
           settings.lottie_format["h"]/2 + get_additional_height()/2]
    anchor = pos
    gen_helpers_transform(lottie["ks"], pos, anchor)

    lottie["ef"].append({})
    gen_effects_fill(lottie["ef"][-1], layer, index.inc())

    lottie["ao"] = settings.LAYER_DEFAULT_AUTO_ORIENT
    lottie["sw"] = settings.lottie_format["w"] + get_additional_width() # Solid Width
    lottie["sh"] = settings.lottie_format["h"] + get_additional_height() # Solid Height

    lottie["sc"] = get_color_hex(layer.get_param("color").get()[0])

    invert = False
    Inv = layer.get_param("invert").get()
    flag = False
    #So far only 'not' convert method seems to be supported for invert param in circle, will add more subsequently.
    if Inv is not None:
        if "bool" not in str(Inv[0]) and "animated" not in str(Inv[0]):
            is_animate = is_animated(Inv[0][0][0])
            flag = True
        else:
            is_animate = is_animated(Inv[0])
        if is_animate == settings.NOT_ANIMATED:
            if flag:
                val = "false"
                if Inv[0][0][0].attrib["value"] == "false":
                    val = "true"
            else:
                val = Inv[0].attrib["value"]
        elif is_animate == settings.SINGLE_WAYPOINT:
            if flag:
                val = "false"
                if Inv[0][0][0][0][0].attrib["value"] == "false":
                    val = "true"
            else:
                val = Inv[0][0][0].attrib["value"]
        else:
            # If animated, always set invert to false
            val = "false"
        if val == "true":
            invert = True

    lottie["ip"] = settings.lottie_format["ip"]
    lottie["op"] = settings.lottie_format["op"]
    lottie["st"] = 0            # Don't know yet
    get_blend(lottie, layer)

    hasMask = True

    lottie["hasMask"] = hasMask
    lottie["masksProperties"] = []
    lottie["masksProperties"].append({})

    if layer.get_type() in {"star", "circle", "rectangle", "filled_rectangle"}:
        bline_point = layer
    else:
        bline_point = layer.get_param("bline", "vector_list")

    gen_mask(lottie["masksProperties"][0], invert, bline_point, index.inc())



# pylint: disable=line-too-long
"""
Store all functions corresponding to pre composition in lottie
"""

import sys
import settings
from sources.precomp import add_precomp_asset
from layers.rotate_layer import gen_layer_rotate
from layers.scale_layer import gen_layer_scale
from layers.translate_layer import gen_layer_translate
sys.path.append("..")


def gen_layer_precomp(lottie, layer, idx):
    """
    Generates a pre-composition layer depending upon the layers inside that
    pre-comp
    Here idx represents the position of layer in the .sif file also

    Args:
        lottie (dict) : Will store the pre-comp layer
        layer  (common.Layer.Layer) : Specifies which layer it is
        idx    (int)  : Index of the layer

    Returns:
        (None)
    """
    lottie["ddd"] = settings.DEFAULT_3D
    lottie["ind"] = idx
    lottie["ty"] = settings.LAYER_PRECOMP_TYPE
    lottie["nm"] = layer.get_description()
    lottie["sr"] = settings.LAYER_DEFAULT_STRETCH
    lottie["ks"] = {}   # Transform properties to be filled

    if layer.get_type() == "rotate":
        # transform properties will be written inside this now
        gen_layer_rotate(lottie["ks"], layer)
        settings.INSIDE_PRECOMP = True
    elif layer.get_type() == "zoom":
        gen_layer_scale(lottie["ks"], layer, "scale_layer")
        settings.INSIDE_PRECOMP = True
    elif layer.get_type() == "translate":
        gen_layer_translate(lottie["ks"], layer)
        settings.INSIDE_PRECOMP = True
    elif layer.get_type() == "stretch":
        gen_layer_scale(lottie["ks"], layer, "stretch_layer")
        settings.INSIDE_PRECOMP = True

    settings.lottie_format["assets"].append({})
    asset = add_precomp_asset(settings.lottie_format["assets"][-1], layer.getparent(), idx)
    lottie["refId"] = asset


    lottie["w"] = settings.lottie_format["w"] + settings.ADDITIONAL_PRECOMP_WIDTH # Experimental increase in width and height of precomposition
    lottie["h"] = settings.lottie_format["h"] + settings.ADDITIONAL_PRECOMP_HEIGHT
    lottie["ao"] = settings.LAYER_DEFAULT_AUTO_ORIENT
    lottie["ip"] = settings.lottie_format["ip"]
    lottie["op"] = settings.lottie_format["op"]
    lottie["st"] = 0            # Don't know yet
    lottie["bm"] = settings.DEFAULT_BLEND    # Always have the default blend



"""
WidthPointList.py
Will store the Parameters class for Width point list
"""

import sys
import common
import properties.shapePropKeyframe as advanced_outline
from synfig.animation import to_Synfig_axis
from common.ActivepointList import ActivepointList
sys.path.append("..")


class WidthPointList:
    """
    Class to store Synfig format Width points list
    """
    def __init__(self, width_point, parent, bline_loop):
        """
        This parent should be parameter
        """
        self.parent = parent
        self.width_point = width_point
        self.bline_loop = bline_loop
        self.type = width_point.attrib["type"]
        self.entry_list = []
        self.extract_entries(self.entry_list)

    def get(self):
        """
        Returns the original param
        """
        return self.width_point

    def __getitem__(self, itr):
        """
        Returns the child corresponding to itr
        """
        return self.entry_list[itr]

    def __setitem__(self, itr, val):
        """
        Sets the value of child corresponding to itr
        """
        self.entry_list[itr] = val

    def get_layer_type(self):
        """
        Recursively go till the top until layer is not reached
        """
        if isinstance(self.parent, common.Layer.Layer):
            return self.parent.get_type()
        return self.parent.get_layer_type()

    def get_layer(self):
        """
        Recursively find the layer of this bline
        """
        if isinstance(self.parent, common.Layer.Layer):
            return self.parent
        return self.parent

    def get_type(self):
        """
        Returns the type of bline
        """
        return self.type

    def get_loop(self):
        """
        Returns whether the bline is looped or not
        """
        loop = False
        if "loop" in self.width_point.keys():
            v = self.width_point.attrib["loop"]
            if v == "true":
                loop = True
        return loop

    def extract_entries(self, entry_list):
        """
        Stores the entries in a list
        """
        for entry in self.width_point:
            # Assuming it's child is always composite for now
            entry_list.append({})
            entry_list[-1]["ActivepointList"] = ActivepointList(entry.attrib.get("on"), entry.attrib.get("off"))
            element = entry[0]
            for param in element:
                tag = param.tag
                if tag == "animated":   # For dynamic list
                    tag = param.attrib["type"]
                entry_list[-1][tag] = common.Param.Param(param, element)
            assert(entry[0].tag == "composite")
            entry_list[-1]["composite"] = element

    def get_entry_list(self):
        """
        Returns the entry list
        """
        return self.entry_list

    def get_len(self):
        """
        Returns the number of entries
        """
        return len(self.entry_list)

    def set_entry(self, itr, entry, tag="vector"):
        """
        Sets the entry at itr to val
        """
        for param in entry:
            self.entry_list[itr][tag] = common.Param.Param(param, entry)

    def get_list_at_frame_2(self, fr):
        """
        Returns list of WidthPoint's at a particular frame
        """
        wplist = []
        for entry in self.get_entry_list():
            pos = entry["position"].get_value(fr)
            pos = to_Synfig_axis(pos, "real")
            width = entry["width"].get_value(fr)
            width = to_Synfig_axis(width, "real")

            side_before = entry["side_before"].get()
            side_before = int(side_before[0].attrib["value"])

            side_after = entry["side_after"].get()
            side_after = int(side_after[0].attrib["value"])

            lower_bound = entry["lower_bound"].get()
            lower_bound = float(lower_bound[0].attrib["value"])

            upper_bound = entry["upper_bound"].get()
            upper_bound = float(upper_bound[0].attrib["value"])

            wplist.append(common.WidthPoint.WidthPoint(pos, width, side_before, side_after, False, lower_bound, upper_bound))

        return wplist
    
    def get_value_at_frame(self, entry, fr):
        pos = entry["position"].get_value(fr)
        pos = to_Synfig_axis(pos, "real")
        width = entry["width"].get_value(fr)
        width = to_Synfig_axis(width, "real")

        side_before = entry["side_before"].get()
        side_before = int(side_before[0].attrib["value"])

        side_after = entry["side_after"].get()
        side_after = int(side_after[0].attrib["value"])

        lower_bound = entry["lower_bound"].get()
        lower_bound = float(lower_bound[0].attrib["value"])

        upper_bound = entry["upper_bound"].get()
        upper_bound = float(upper_bound[0].attrib["value"])

        curr = common.WidthPoint.WidthPoint(pos, width, side_before, side_after, False, lower_bound, upper_bound)
        return curr

    def get_list_at_frame(self, fr):
        """
        Refer: https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/valuenodes/valuenode_wplist.cpp#L281
        """
        wplist = []
        rising = [False]

        for entry in self.get_entry_list():
            amount = entry["ActivepointList"].amount_at_time(fr, rising)
            assert(amount >= 0)
            assert(amount <= 1)

            curr = self.get_value_at_frame(entry, fr)

            # it's fully on
            if amount > 1 - 0.0000001:
                wplist.append(curr)
            # it's partly on
            elif amount > 0:
                # This is where the interesting stuff happens
                # on_time, and off_time are never needed, so why bother writing
                # them?

                i_width = self.interpolated_width(curr.get_norm_position(self.get_loop()), fr)
                curr_width = curr.get_width()

                curr.set_width(i_width*(1.0-amount)+(curr_width)*amount)
                wplist.append(curr)
        return wplist

    def interpolated_width(self, position, time):
        prev = self.find_prev_valid_entry_by_position(position, time)
        neext = self.find_next_valid_entry_by_position(position, time)
        prev.normalize(self.get_loop())
        neext.normalize(self.get_loop())
        return advanced_outline.advanced_outline.widthpoint_interpolate(prev, neext, position)

    def find_prev_valid_entry_by_position(self, position, time):
        prev_pos = -123456
        prev_ret = common.WidthPoint.WidthPoint(prev_pos, 0.0, 0, 0, False)
        if len(self.get_entry_list()) == 0:
            return prev_ret
        for entry in self.get_entry_list():
            curr = self.get_value_at_frame(entry, time)
            curr_pos = curr.get_norm_position(self.get_loop())

            status = entry["ActivepointList"].status_at_time(time)
            if (curr_pos < position) and (curr_pos > prev_pos) and status:
                prev_pos = curr_pos
                prev_ret = curr

        if prev_ret.get_position() == -123456:
            if self.bline_loop:
                prev_ret = self.find_prev_valid_entry_by_position(2.0, time)
            else:
                prev_ret = self.find_next_valid_entry_by_position(-1.0, time)
                prev_ret.set_position(0.0)

        return prev_ret

    def find_next_valid_entry_by_position(self, position, time):
        next_pos = 1.0
        next_ret = common.WidthPoint.WidthPoint(next_pos, 0.0, 0, 0, False)

        for entry in self.get_entry_list():
            curr = self.get_value_at_frame(entry, time)
            curr_pos = curr.get_norm_position(self.get_loop())
            status = entry["ActivepointList"].status_at_time(time)
            if (curr_pos > position) and (curr_pos < next_pos) and status:
                next_pos = curr_pos
                next_ret = curr
        return next_ret



#!/usr/bin/env python

#
# Copyright (c) 2012 by Konstantin Dmitriev <ksee.zelgadis@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys

def process(filename):
	
	# Read the input file
	inputfile_f = open(filename, 'r')
	inputfile_contents = inputfile_f.readlines()
	inputfile_f.close()
	
	# Now write results to the same file
	inputfile_f = open(filename, 'w')

	for line in inputfile_contents:
		if "<layer " in line:
			inputfile_f.write(line.replace(' active="false" ',' active="true" '))
		else:
			inputfile_f.write(line)
	inputfile_f.close()

if len(sys.argv) < 2:
	sys.exit()
else:
	process(sys.argv[1])




# pylint: disable=line-too-long
"""
misc.py
Some miscellaneous functions and classes will be provided here
"""

import sys
import math
import settings
from common.Vector import Vector
from common.Color import Color
from common.Gradient import Gradient
sys.path.append("..")


def approximate_equal(a, b):
    """
    Need to define this function somewhere else, a and b are of type "float"

    Args:
        a (float) : First number to be compared
        b (float) : Second number to be compared

    Returns:
        (bool) : True if the numbers are approximately equal under precision
               : False otherwise
    """
    precision = 1e-8
    if a < b:
        return b - a < precision
    return a - b < precision


def real_high_precision():
    """
    Synfig format real high precision
    https://github.com/synfig/synfig/blob/ae11655a9bba068543be7a5df9090958579de78e/synfig-core/src/synfig/real.h#L55

    Args:
        (None)

    Returns:
        (float) : The value of real_high_precision as described in Synfig
    """
    return 1e-10


def calculate_pixels_per_unit():
    """
    Gives the value of 1 unit in terms of pixels according to the canvas defined

    Args:
        (None)

    Returns:
        (float) : Pixels per unit
    """
    image_width = float(settings.lottie_format["w"])
    image_area_width = settings.view_box_canvas["val"][2] - settings.view_box_canvas["val"][0]
    settings.PIX_PER_UNIT = image_width / image_area_width
    return settings.PIX_PER_UNIT


def change_axis(x_val, y_val, is_transform=True):
    """
    Convert synfig axis coordinates into lottie format

    Args:
        x_val (float | str) : x axis value in pixels
        y_val (float | str) : y axis value in pixels
        is_transform (`obj`: bool, optional) : Is this value used in transform module?

    Returns:
        (list)  : x and y axis value in Lottie format
    """
    x_val, y_val = float(x_val), float(y_val)
    if is_transform:
        x_val, y_val = x_val, -y_val
    else:
        x_val, y_val = x_val + settings.lottie_format["w"]/2, -y_val + settings.lottie_format["h"]/2
    return [x_val, y_val]


def parse_position(animated, i):
    """
    To convert the synfig coordinates from units(initially a string) to pixels
    Depends on whether a vector is provided to it or a real value
    If real value is provided, then time is also taken into consideration

    Args:
        animated (lxml.etree._Element) : Stores animation which contains waypoints
        i        (int)                 : Iterator over animation

    Returns:
        (common.Vector.Vector) If the animated type is not color
        (common.Color.Color)  Else if the animated type is color
    """
    if animated.attrib["type"] == "vector":
        pos = [float(animated[i][0][0].text),
               float(animated[i][0][1].text)]
        pos = [settings.PIX_PER_UNIT*x for x in pos]

    elif animated.attrib["type"] in {"real", "circle_radius"}:
        pos = parse_value(animated, i)

    elif animated.attrib["type"] == "angle":
        pos = [get_angle(float(animated[i][0].attrib["value"])),
               get_frame(animated[i])]

    elif animated.attrib["type"] == "base":
        pos = [float(animated[i][0].attrib["value"])*settings.PIX_PER_UNIT,
               get_frame(animated[i])]

    elif animated.attrib["type"] in {"composite_convert", "region_angle", "star_angle_new", "scalar_multiply"}:
        pos = [float(animated[i][0].attrib["value"]),
               get_frame(animated[i])]

    elif animated.attrib["type"] == "rotate_layer_angle":
        # Angle needs to made neg of what they are
        pos = [-float(animated[i][0].attrib["value"]),
               get_frame(animated[i])]

    elif animated.attrib["type"] == "opacity":
        pos = [float(animated[i][0].attrib["value"]) * settings.OPACITY_CONSTANT,
               get_frame(animated[i])]

    elif animated.attrib["type"] == "effects_opacity":
        pos = [float(animated[i][0].attrib["value"]),
               get_frame(animated[i])]

    elif animated.attrib["type"] == "points":
        pos = [int(animated[i][0].attrib["value"]) * settings.PIX_PER_UNIT,
               get_frame(animated[i])]

    elif animated.attrib["type"] == "bool":
        val = animated[i][0].attrib["value"]
        if val == "false":
            val = 0
        else:
            val = 1
        pos = [val, get_frame(animated[i])]

    elif animated.attrib["type"] == "rectangle_size":
        pos = parse_value(animated, i)
        vec = Vector(pos[0], pos[1], animated.attrib["type"])
        vec.add_new_val(float(animated[i][0].attrib["value2"]) * settings.PIX_PER_UNIT)
        return vec

    elif animated.attrib["type"] == "image_scale":
        val = float(animated[i][0].attrib["value"])
        val2 = get_frame(animated[i])
        vec = Vector(val, val2, animated.attrib["type"])
        vec.add_new_val(float(animated[i][0].attrib["value2"]))
        return vec

    elif animated.attrib["type"] == "scale_layer_zoom":
        val = (math.e ** float(animated[i][0].attrib["value"])) * 100
        val2 = get_frame(animated[i])
        vec = Vector(val, val2, animated.attrib["type"])
        vec.add_new_val(val)
        return vec

    elif animated.attrib["type"] == "stretch_layer_scale":
        val1 = float(animated[i][0][0].text) * 100
        val3 = float(animated[i][0][1].text) * 100
        vec = Vector(val1, get_frame(animated[i]), animated.attrib["type"])
        vec.add_new_val(val3)
        return vec

    elif animated.attrib["type"] in {"group_layer_scale","blur_anim_x","blur_anim_y"}:
        if animated.attrib["type"] == "group_layer_scale":
            val1 = float(animated[i][0][0].text) * 100
            val3 = float(animated[i][0][1].text) * 100
            vec = Vector(val1, get_frame(animated[i]), animated.attrib["type"])
            vec.add_new_val(val3)
        elif animated.attrib["type"] == "blur_anim_x":
            val1 = float(animated[i][0][0].text) * 100
            vec = Vector(val1, get_frame(animated[i]), animated.attrib["type"])
        else:
            val1 = float(animated[i][0][1].text) * 100
            vec = Vector(val1, get_frame(animated[i]), animated.attrib["type"])
        return vec

    elif animated.attrib["type"] == "time":
        val = parse_time(animated[i][0].attrib["value"])    # Needed in seconds
        val2 = get_frame(animated[i])   # Needed in frames
        return Vector(val, val2, animated.attrib["type"])

    elif animated.attrib["type"] == "color":
        red = float(animated[i][0][0].text)
        green = float(animated[i][0][1].text)
        blue = float(animated[i][0][2].text)
        alpha = float(animated[i][0][3].text)
        red = red ** (1/settings.GAMMA[0])
        green = green ** (1/settings.GAMMA[1])
        blue = blue ** (1/settings.GAMMA[2])
        return Color(red, green, blue, alpha)
    
    elif animated.attrib["type"] == "gradient":
        return Gradient(animated[i][0])

    return Vector(pos[0], pos[1], animated.attrib["type"])


def parse_value(animated, i):
    """
    To convert the synfig value parameter from units to pixels
    and also take into consideration the time parameter

    Args:
        animated (lxml.etree._Element) : Stores animation which holds waypoints
        i        (int)                 : Iterator for animation

    Returns:
        (list)  : [value, time] is returned
    """
    pos = [float(animated[i][0].attrib["value"]) * settings.PIX_PER_UNIT,
           get_frame(animated[i])]
    return pos


def get_angle(theta):
    """
    Converts the .sif angle into lottie angle
    .sif uses positive x-axis as the start point and goes anticlockwise
    lottie uses positive y-axis as the start point and goes clockwise

    Args:
        theta (float) : Stores Synfig format angle

    Returns:
        (int)   : Lottie format angle
    """
    theta = int(theta)
    shift = -int(theta / 360)
    theta = theta % 360
    theta = (90 - theta) % 360

    theta = theta + shift * 360
    return theta


def is_animated(node):
    """
    Tells whether a parater is animated or not

    Args:
        node (lxml.etree._Element) : Animation which stores waypoints

    Returns:
        (int) : Depending upon whether the parameter is animated, following
                values are returned
                0: If not animated
                1: If only single waypoint is present
                2: If more than one waypoint is present
    """
    case = settings.NOT_ANIMATED
    if node.tag == "animated":
        if len(node) == 1:
            case = settings.SINGLE_WAYPOINT
        else:
            case = settings.ANIMATED

    return case


def clamp_col(color, gamma):
    """
    This function converts the colors into int and takes them to the range of
    0-255

    Args:
        color (float) : Synfig format color value

    Returns:
        (int) : Color value between 0-255
    """
    color = color ** (1/gamma)
    color *= 255
    color = int(color)
    return max(0, min(color, 255))


def get_color_hex(node):
    """
    Convert the <color></color> from rgba to hex format

    Args:
        node (lxml.etree._Element) : Synfig format color parameter

    Returns:
        (str) : hex format of color
    """
    red, green, blue = 1, 0, 0
    for col in node:
        if col.tag == "r":
            red = float(col.text)
        elif col.tag == "g":
            green = float(col.text)
        elif col.tag == "b":
            blue = float(col.text)
    # Convert to 0-255 range
    red, green, blue = clamp_col(red, settings.GAMMA[0]), clamp_col(green, settings.GAMMA[1]), clamp_col(blue, settings.GAMMA[2])

    # https://stackoverflow.com/questions/3380726/converting-a-rgb-color-tuple-to-a-six-digit-code-in-python/3380739#3380739
    ret = "#{0:02x}{1:02x}{2:02x}".format(red, green, blue)
    return ret


def get_frame(waypoint):
    """
    Given a waypoint, it parses the time to frames

    Args:
        waypoint (lxml.etree._Element) : Synfig format waypoint

    Returns:
        (int) : the frame at which waypoint is present
    """
    time = get_time(waypoint)
    frame = time * settings.lottie_format["fr"]
    frame = round(frame)
    return frame


def get_time(waypoint):
    """
    Given a waypoint, it parses the string time to float time

    Args:
        waypoint (lxml.etree._Element) : Synfig format waypoint

    Returns:
        (float) : the time in seconds at which the waypoint is present
    """
    return parse_time(waypoint.attrib["time"])


def parse_time(time_in_str):
    """
    Given a string, it parses time to float time

    Args:
        time_in_str (str) : Time in string format

    Returns:
        (float) : the time in seconds at represented by the string
    """
    time = time_in_str.split(" ")
    final = 0
    for frame in time:
        if frame[-1] == "h":
            final += float(frame[:-1]) * 60 * 60
        elif frame[-1] == "m":
            final += float(frame[:-1]) * 60
        elif frame[-1] == "s":
            final += float(frame[:-1])
        elif frame[-1] == "f":  # This should never happen according to my code
            raise ValueError("In waypoint, time is never expected in frames.")
    return final


def get_vector(waypoint):
    """
    Given a waypoint, it parses the string vector into Vector class defined in
    this converter

    Args:
        waypoint (lxml.etree._Element) : Synfig format waypoint

    Returns:
        (common.Vector.Vector) : x and y axis values stores in Vector format
    """
    # converting radius and angle to a vector
    if waypoint.tag == "radial_composite":
        for child in waypoint:
            if child.tag == "radius":
                radius = float(child[0].attrib["value"])
                radius *= settings.PIX_PER_UNIT
            elif child.tag == "theta":
                angle = float(child[0].attrib["value"])
        x, y = radial_to_tangent(radius, angle)
    else:
        x = float(waypoint[0][0].text)
        y = float(waypoint[0][1].text)
    return Vector(x, y)


def radial_to_tangent(radius, angle):
    """
    Converts a tangent from radius and angle format to x, y axis coordinate
    system

    Args:
        radius (float) : radius of a vector
        angle  (float) : angle in degrees

    Returns:
        (float) : The x-coordinate of the vector
        (float) : The y-coordinate of the vector
    """
    angle = math.radians(angle)
    x = radius * math.cos(angle)
    y = radius * math.sin(angle)
    return x, y


def set_vector(waypoint, pos):
    """
    Given a waypoint and pos(Vector), it set's the waypoint's vectors

    Args:
        waypoint (lxml.etree._Element) : Synfig format waypoint

    Returns:
        (None)
    """
    waypoint[0][0].text = str(pos.val1)
    waypoint[0][1].text = str(pos.val2)


def modify_final_dump(obj):
    """
    This function will remove unwanted keys from the final dictionary and also
    modify the floats according to our precision

    Args:
        obj (float | dict | list | tuple) : The object to be modified

    Returns:
        (float | dict | list) : Depending upon arguments, obj type is decided
    """
    if isinstance(obj, float):
        return round(obj, settings.FLOAT_PRECISION)
    elif isinstance(obj, dict):
        return dict((k, modify_final_dump(v)) for k, v in obj.items() if k not in ["synfig_i", "synfig_o"])
    elif isinstance(obj, (list, tuple)):
        return list(map(modify_final_dump, obj))
    return obj



# pylint: disable=line-too-long
"""
Python plugin to convert the .sif format into lottie json format
input   : FILE_NAME.sif
output  : FILE_NAME.json
        : FILE_NAME.html
        : FILE_NAME.log

Supported Layers are mentioned below
"""
import os
import json
import sys
import logging
from lxml import etree
from canvas import gen_canvas
from layers.driver import gen_layers
from common.misc import modify_final_dump
from common.Canvas import Canvas
import settings
import argparse


def parse(file_name):
    """
    Driver function for parsing .sif to lottie(.json) format

    Args:
        file_name (str) : Synfig file name that needs to be parsed to Lottie format

    Returns:
        (str) : File name in json format
    """
    tree = etree.parse(file_name)
    root = tree.getroot()  # canvas
    gen_canvas(settings.lottie_format, root)

    # Storing the file name
    settings.file_name["fn"] = file_name

    # Storing the file directory
    settings.file_name["fd"] = os.path.dirname(file_name)

    # Initialize the logging
    init_logs()

    settings.lottie_format["layers"] = []
    canvas = Canvas(root)
    gen_layers(settings.lottie_format["layers"], canvas, canvas.get_num_layers() - 1)

    return json.dumps(modify_final_dump(settings.lottie_format))


def gen_html(file_name):
    """
    Generates an HTML file which will allow end user to easily playback
    animation in a web browser

    Args:
        file_name (str) : Stores the HTML file name

    Returns:
        (None)
    """
    if len(settings.blur_dictionary) != 0 or settings.OUTLINE_FLAG:
        settings.lottie_format["v"] = "5.6.5"
        bodymovin_path = os.path.join(os.path.dirname(sys.argv[0]), "test_bodymovin.js")
    else:
        bodymovin_path = os.path.join(os.path.dirname(sys.argv[0]), "bodymovin.js")

    with open(bodymovin_path, "r") as f:
        bodymovin_script = f.read()

    html_text = \
"""<html xmlns="http://www.w3.org/1999/xhtml">
<meta charset="UTF-8">
<head>
    <style>
        body{{
            background-color:#fff;
            margin: 0px;
            height: 100%;
            overflow: hidden;
        }}
        #lottie{{
            background-color:#fff;
            width:100%;
            height:100%;
            display:block;
            overflow: hidden;
            transform: translate3d(0,0,0);
            text-align: center;
            opacity: 1;
        }}

    </style>
</head>
<body>

{bodymovin_script}

<div id="lottie"></div>
<script>
    var animationData = {file_name_data};
    var params = {{
        container: document.getElementById('lottie'),
        renderer: 'svg',
        loop: true,
        autoplay: true,
        animationData: animationData
    }};

    var anim;

    anim = lottie.loadAnimation(params);

</script>
</body>
</html>
"""
    return html_text.format(bodymovin_script=bodymovin_script, file_name_data=json.dumps(modify_final_dump(settings.lottie_format)))


def init_logs():
    """
    Initializes the logger, sets the level of the logging(DEBUG | INFO : depending on what is
    specified)
    """
    logging.basicConfig(stream=sys.stdout, format='%(name)s - %(levelname)s - %(message)s')
    logging.getLogger().setLevel(logging.DEBUG)


parser = argparse.ArgumentParser()
parser.add_argument("infile")
parser.add_argument("outfile")
ns = parser.parse_args()
	
settings.init()
settings.WITHOUT_VARIABLE_WIDTH = True

out = parse(ns.infile)
if ns.outfile.endswith(".html"):
    out = gen_html(out)

with open(ns.outfile, "w") as fil:
    fil.write(out)



"""
This module will store the expression controllers of lottie/AE
"""

import sys
import settings
from effects.slider import gen_effects_slider
from effects.point import gen_effects_point
sys.path.append("../")


def gen_effects_controller(lottie, value, anim_type):
    """
    Generates the dictionary correspondingt to effects/controller.json
    """
    lottie["ty"] = settings.EFFECTS_CONTROLLER  # Effect type
    idx = settings.controller_count.inc()
    lottie["nm"] = "Controller" + str(idx)
    lottie["ix"] = idx

    lottie["ef"] = []
    lottie["ef"].append({})
    if anim_type in {"vector", "group_layer_scale", "stretch_layer_scale", "circle_radius"}:
        gen_effects_point(lottie["ef"][-1], value, idx)
    else:
        gen_effects_slider(lottie["ef"][-1], value, idx)



# pylint: disable=line-too-long
"""
Will store all the functions related to modify and play with group layer of
Synfig/precomp layer of Lottie
"""

import sys
import settings
from common.misc import is_animated
from common.Vector import Vector
import common
sys.path.append("..")


def get_offset():
    """
    Computes the offset by which the layers need to be moved with

    Args:
        (None)
    Returns:
        (common.Vector.Vector) : offset
    """
    x = settings.ADDITIONAL_PRECOMP_WIDTH / 2
    y = settings.ADDITIONAL_PRECOMP_HEIGHT / 2
    offset = Vector(x, -y)
    offset /= settings.PIX_PER_UNIT
    return offset


def update_layer(node):
    """
    Inserts necassary offset in the positions of the layers if they lie inside
    another composition of Lottie

    Args:
        node (common.Layer.Layer | lxml.etree._Element) : Can be a layer or parameter of a layer

    Returns:
        (None)
    """
    # This if condition is not applicable for group, rotate, precomp... layers
    if not settings.INSIDE_PRECOMP:
        return

    update_dict = []
    if isinstance(node, common.Layer.Layer):
        compare = {"center", "origin", "point1", "point2", "tl", "br"}
        for param in compare:
            get = node.get_param(param).get()
            if get is not None:
                update_dict.append(get)
    else:
        update_dict.append(node)

    for param in update_dict:
        update_pos(param)


def update_pos(origin):
    """
    Updates the position of parameter in Synfig format

    Args:
        origin (lxml.etree._Element) : Stores the position of element

    Returns:
        (None)
    """
    offset = get_offset()
    is_animate = is_animated(origin[0])
    if is_animate == settings.NOT_ANIMATED:
        add(origin[0], offset)
    else:
        for waypoint in origin[0]:
            add(waypoint[0], offset)


def add(vector, offset):
    """
    Helper function to modify Synfig xml

    Args:
        vector (lxml.etree._Element) : Position in Synfig format
        offset (common.Vector.Vector) : offset to be added to that position

    Returns:
        (None)
    """
    vector[0].text = str(float(vector[0].text) + offset[0])
    vector[1].text = str(float(vector[1].text) + offset[1])


def get_additional_width():
    """
    Returns the increase in pre-composition width depending on whether we are
    inside a pre-composition yet or not

    Args:
        (None)
    Returns:
        (float | int) : the additional width
    """
    if settings.INSIDE_PRECOMP:
        return settings.ADDITIONAL_PRECOMP_WIDTH
    return 0


def get_additional_height():
    """
    Returns the increase in pre-composition height depending on whether we are
    inside a pre-composition yet or not

    Args:
        (None)
    Returns:
        (float | int) : the additional height
    """
    if settings.INSIDE_PRECOMP:
        return settings.ADDITIONAL_PRECOMP_HEIGHT
    return 0

# Should be -x, +y -> the offset



# pylint: disable=line-too-long
"""
Will store all the functions and modules for generation of rectangle layer
in Lottie format
"""

import sys
from lxml import etree
from common.misc import approximate_equal
from common.Vector import Vector
from synfig.animation import to_Synfig_axis
from properties.shapePropKeyframe.helper import add, insert_dict_at, quadratic_to_cubic
sys.path.append("../../")


def gen_list_rectangle(lottie, layer):
    """
    Generates a shape layer corresponding to rectangle layer by manipulating the
    parameters of the rectangle

    Args:
        lottie (dict) : Lottie format rectangle layer will be stored in this
        layer  (common.Layer.Layer) : Synfig format rectangle layer

    Returns:
        (None)
    """
    ################### SECTION 1 #########################
    # Inserting waypoints if not animated and finding the first and last frame
    window = {}
    window["first"] = sys.maxsize
    window["last"] = -1

    point1 = layer.get_param("point1")
    point2 = layer.get_param("point2")
    expand = layer.get_param("expand")
    bevel = layer.get_param("bevel")
    bevCircle = layer.get_param("bevCircle")

    if expand.get() is None:    # Means filled rectangle layer
        st = "<param name='expand'><real value='0.0'/></param>"
        expand = etree.fromstring(st)
        expand = layer.add_param("expand", expand)

    if bevel.get() is None:     # For rectangle layer in stable version 1.2.2
        st = "<param name='bevel'><real value='0.0'/></param>"
        bevel = etree.fromstring(st)
        bevel = layer.add_param("bevel", bevel)
        st = "<param name='bevCircle'><bool value='false'/></param>"
        bevCircle = etree.fromstring(st)
        bevCircle = layer.add_param("bevCircle", bevCircle)

    # Animating point1
    point1.update_frame_window(window)
    point1.animate("vector")

    # Animating point2
    point2.update_frame_window(window)
    point2.animate("vector")

    # Animating expand
    expand.update_frame_window(window)
    expand.animate("real")

    # Animating bevel
    bevel.update_frame_window(window)
    bevel.animate("real")

    # Animating bevCircle
    bevCircle.update_frame_window(window)
    bevCircle.animate_without_path("bool")

    # Minimizing the window size
    if window["first"] == sys.maxsize and window["last"] == -1:
        window["first"] = window["last"] = 0
    ################# END OF SECTION 1 ###################

    ################ SECTION 2 ###########################
    # Generating values for all the frames in the window

    fr = window["first"]
    while fr <= window["last"]:
        st_val, en_val = insert_dict_at(lottie, -1, fr, False)

        synfig_rectangle(st_val, point1, point2, expand, bevel, bevCircle, fr)
        synfig_rectangle(en_val, point1, point2, expand, bevel, bevCircle, fr + 1)

        fr += 1
    # Setting the final time
    lottie.append({})
    lottie[-1]["t"] = fr


def synfig_rectangle(st_val, point1_p, point2_p, expand_p, bevel_p, bevCircle, fr):
    """
    Calculates the points for the rectangle layer as in Synfig:
    https://github.com/synfig/synfig/blob/678cc3a7b1208fcca18c8b54a29a20576c499927/synfig-core/src/modules/mod_geometry/rectangle.cpp

    Args:
        st_val (dict) : Lottie format rectangle will be stored in this
        point1_p (self.Param.Param) : Lottie format point1 animation
        point2_p (self.Param.Param) : Lottie format point2 animation
        expand_p (self.Param.Param) : Lottie format expand parameter animation
        bevel_p (self.Param.Param) : Lottie format bevel parameter animation
        bevCircle (lxml.etree._Element) : Animation of bevCircle in Synfig format
        fr (int) : Frame Number

    Returns:
        (None)
    """

    expand = abs(to_Synfig_axis(expand_p.get_value(fr), "real"))
    bevel = abs(to_Synfig_axis(bevel_p.get_value(fr), "real"))
    p0 = to_Synfig_axis(point1_p.get_value(fr), "vector")
    p0 = Vector(p0[0], p0[1])
    p1 = to_Synfig_axis(point2_p.get_value(fr), "vector")
    p1 = Vector(p1[0], p1[1])
    if p1[0] < p0[0]:
        p0[0], p1[0] = p1[0], p0[0]
    if p1[1] < p0[1]:
        p0[1], p1[1] = p1[1], p0[1]

    bev_circle = bevCircle.get_value(fr)

    w = p1[0] - p0[0] + 2*expand
    h = p1[1] - p0[1] + 2*expand
    bev = bevel
    if bevel > 1:
        bev = 1
    if bev_circle:
        bevx = min(w*bev/2.0, h*bev/2.0)
        bevy = min(w*bev/2.0, h*bev/2.0)
    else:
        bevx = w*bev/2.0
        bevy = h*bev/2.0

    # Setup chunk list
    chunk_list = []

    if approximate_equal(bevel, 0.0):
        chunk_list.append([Vector(p0[0] - expand, p0[1] - expand), Vector(), Vector()])
        chunk_list.append([Vector(p1[0] + expand, p0[1] - expand), Vector(), Vector()])
        chunk_list.append([Vector(p1[0] + expand, p1[1] + expand), Vector(), Vector()])
        chunk_list.append([Vector(p0[0] - expand, p1[1] + expand), Vector(), Vector()])
    else:
        cur = Vector(p1[0] + expand - bevx, p0[1] - expand)
        chunk_list.append([cur, Vector(), Vector()])
        prev = cur

        cur = Vector(p1[0] + expand, p0[1] - expand + bevy)
        cp1, cp2 = quadratic_to_cubic(cur, Vector(p1[0] + expand, p0[1] - expand), prev)
        chunk_list[-1][2] = cp2 - prev
        chunk_list.append([cur, cur - cp1, Vector()])
        prev = cur

        cur = Vector(p1[0] + expand, p1[1] + expand - bevy)
        chunk_list.append([cur, Vector(), Vector()])
        prev = cur

        cur = Vector(p1[0] + expand - bevx, p1[1] + expand)
        cp1, cp2 = quadratic_to_cubic(cur, Vector(p1[0] + expand, p1[1] + expand), prev)
        chunk_list[-1][2] = cp2 - prev
        chunk_list.append([cur, cur - cp1, Vector()])
        prev = cur

        cur = Vector(p0[0] - expand + bevx, p1[1] + expand)
        chunk_list.append([cur, Vector(), Vector()])
        prev = cur

        cur = Vector(p0[0] - expand, p1[1] + expand - bevy)
        cp1, cp2 = quadratic_to_cubic(cur, Vector(p0[0] - expand, p1[1] + expand), prev)
        chunk_list[-1][2] = cp2 - prev
        chunk_list.append([cur, cur - cp1, Vector()])
        prev = cur

        cur = Vector(p0[0] - expand, p0[1] - expand + bevy)
        chunk_list.append([cur, Vector(), Vector()])
        prev = cur

        cur = Vector(p0[0] - expand + bevx, p0[1] - expand)
        cp1, cp2 = quadratic_to_cubic(cur, Vector(p0[0] - expand, p0[1] - expand), prev)
        chunk_list[-1][2] = cp2 - prev
        chunk_list.append([cur, cur - cp1, Vector()])
        prev = cur

    add(chunk_list, st_val, [0, 0], True)



"""
Will store all the functions corresponding to Image Assets in lottie
"""

import sys
import settings
import layers.driver
sys.path.append("..")


def add_precomp_asset(lottie, canvas, layer_itr):
    """
    Generates a unique ID for precomp asset and returns it to the caller
    function

    Args:
        lottie (dict) : Asset will be stored here in lottie format
        cavas (common.Canvas.Canvas) : Canvas for which layers are to be generated
        layer_itr (int) : Index of the layer in canvas

    Returns:
        (str) : Unique ID of the asset
    """
    lottie["id"] = "precomp_" + str(settings.num_precomp.inc())
    lottie["layers"] = []   # If no layer is added, then might result in an error, keep in mind

    # Parsing the canvas to class canvas
    layers.driver.gen_layers(lottie["layers"], canvas, layer_itr-1)
    return lottie["id"]



# pylint: disable=line-too-long
"""
Will store all the functions corresponding to Blur Layer in lottie
"""

import settings
from properties.valueKeyframed import gen_value_Keyframed
from common.misc import is_animated
from properties.value import gen_properties_value
from common.Count import Count

def blurriness(lottie,parameter,idx,direction):
	"""
	This function will be called for blur layer direction to create the blurriness
	value dictionary

	Args:
		lottie    (dict)               : Lottie Dictionary for blurriness
		parameter (common.Param.Param) : Synfig format param for blur size
		idx       (int)                : Stores the index(number of) of blurriness
		direction (string)             : Indicates the direction of blurring

	Returns:
		(None)
	"""
	lottie["ty"] =  0
	lottie["nm"] = "Blurriness"
	lottie["mn"] = "Gaussian Blur 1"
	lottie["ix"] = idx
	lottie["v"]  = {}
	is_animate = is_animated(parameter[0])
	if is_animate == settings.ANIMATED:
		if direction == "horizontal":
			parameter[0].attrib['type'] = 'blur_anim_x'
			gen_value_Keyframed(lottie["v"],parameter[0],1)
		else:
			parameter[0].attrib['type'] = 'blur_anim_y'
			gen_value_Keyframed(lottie["v"],parameter[0],1)
	else:
		if is_animate == settings.NOT_ANIMATED:
			if direction == "horizontal":
				val = float(parameter[0][0].text) * 100
			else:
				val = float(parameter[0][1].text) * 100
		else:
			if direction == "horizontal":
				val = float(parameter[0][0][0].text) * 100
			else:
				val = float(parameter[0][0][1].text) * 100
		gen_properties_value(lottie["v"],
							 val,
							 1,
							 settings.DEFAULT_ANIMATED,
							 settings.NO_INFO)

def generate_dimensions_dict(lottie,idx,direction):
	"""
	This function will be called for filling blur layer properties

	Args:
		lottie    (dict)               : Lottie Dictionary for blur layer properties
		idx       (int)                : Stores the index(number of) of blur layer properties
		direction (string)             : Indicates the direction of blurring

	Returns:
		(None)
	"""
	lottie["ty"] =  7
	lottie["nm"] = "Blur Dimensions"
	lottie["mn"] = "Gaussian Blur 2"
	lottie["ix"] = idx
	lottie["v"]  = {}
	if direction == "horizontal":
		lottie["nm"] = "Blur Dimensions"
		lottie["mn"] = "Gaussian Blur 2"
		gen_properties_value(lottie["v"],
							 2,
							 idx,
							 settings.DEFAULT_ANIMATED,
							 settings.NO_INFO)
	elif direction == "vertical":
		lottie["nm"] = "Blur Dimensions"
		lottie["mn"] = "Gaussian Blur 2"
		gen_properties_value(lottie["v"],
							 3,
							 idx,
							 settings.DEFAULT_ANIMATED,
							 settings.NO_INFO)
	else:
		lottie["nm"] = "Repeat Edge Pixels"
		lottie["mn"] = "Gaussian Blur 3"
		gen_properties_value(lottie["v"],
							 0,
							 idx,
							 settings.DEFAULT_ANIMATED,
							 settings.NO_INFO)
def fill_blur_dict(lottie, layer, idx,direction):
	"""
	This function will be called for each blur layer separately for the two directions

	Args:
		lottie    (dict)               : Lottie Dictionary for blur layers
		layers    (common.Layer.Layer) : Synfig format layer
		idx       (int)                : Stores the index(number of) of blur layer
		direction (string)             : Indicates the direction of blurring

	Returns:
		(None)
	"""
	index = Count()
	lottie["ty"] = settings.BLUR_TYPE
	lottie["nm"] = layer.get_description()
	lottie["mn"] = lottie["nm"]
	lottie["np"] = 5
	lottie["en"] = 1
	lottie["ix"] = idx
	lottie["ef"] = []
	size = layer.get_param("size").get()
	temp_blurriness = {}
	blurriness(temp_blurriness,size,index.inc(),direction)
	temp_directions = {}
	generate_dimensions_dict(temp_directions,index.inc(),direction)
	temp_pixels = {}
	generate_dimensions_dict(temp_pixels,index.inc(),"edge_pixels")

	lottie["ef"].append(temp_blurriness)
	lottie["ef"].append(temp_directions)
	lottie["ef"].append(temp_pixels)

def gen_layer_blur(lottie, layers):
	"""
	This function will be called for each canvas/composition. Main function to
	generate all the layers

	Args:
		lottie (dict) : Lottie Dictionary for blur layers
		layers (List) : Dictionary of Synfig format layers

	Returns:
		(None)
	"""
	index = Count()
	for layer in layers:
		blur_dict_x = {}
		fill_blur_dict(blur_dict_x,layer,index.inc(),"horizontal")
		blur_dict_y = {}
		fill_blur_dict(blur_dict_y,layer,index.inc(),"vertical")
		lottie.append(blur_dict_x)
		lottie.append(blur_dict_y)



# pylint: disable=line-too-long
"""
Will store all the functions and modules for generation of region layer
in Lottie format
"""

import sys
import settings
from common.Bline import Bline
from properties.shapePropKeyframe.helper import insert_dict_at, animate_tangents, convert_tangent_to_lottie
from properties.shapePropKeyframe.outline import equalize_length
from synfig.animation import to_Lottie_axis
sys.path.append("../../")


def gen_bline_region(lottie, bline_point):
    """
    Generates the dictionary corresponding to properties/shapePropKeyframe.json,
    given a bline/spline

    Args:
        lottie     (dict) : Lottie generated keyframes will be stored here for shape/path
        bline_path (common.Param.Param) : shape/path store in Synfig format

    Returns:
        (None)
    """
    ################### SECTION 1 #########################
    # Inserting waypoints if not animated and finding the first and last frame
    window = {}
    window["first"] = sys.maxsize
    window["last"] = -1

    bline = Bline(bline_point[0], bline_point)
    loop = bline.get_loop()

    for entry in bline.get_entry_list():
        pos = entry["point"]
        origin = entry["origin"]
        t1 = entry["t1"]
        t2 = entry["t2"]
        split_r = entry["split_radius"]
        split_a = entry["split_angle"]
        width = entry["width"]  # Not needed, but required by Bline class
        # Hence we do not update the window also
        width.animate("real")

        origin.update_frame_window(window)
        origin.animate("real")

        # Necassary to update this before inserting new waypoints, as new
        # waypoints might include there on time: 0 seconds
        pos.update_frame_window(window)
        pos.animate("vector", True)

        split_r.update_frame_window(window)
        split_r.animate_without_path("bool")

        split_a.update_frame_window(window)
        split_a.animate_without_path("bool")

        animate_tangents(t1, window)
        animate_tangents(t2, window)

        entry["ActivepointList"].update_frame_window(window)

    layer = bline.get_layer().get_layer()
    origin = layer.get_param("origin")

    # Animating the origin
    origin.update_frame_window(window)
    origin.animate("vector")

    # Minimizing the window size
    if window["first"] == sys.maxsize and window["last"] == -1:
        window["first"] = window["last"] = 0
    ################# END OF SECTION 1 ###################

    ################ SECTION 2 ###########################
    # Generating values for all the frames in the window
    fr = window["first"]
    lottie_st_list = []
    lottie_en_list = []
    while fr <= window["last"]:
        st_val, en_val = insert_dict_at(lottie, -1, fr, loop)
        lottie_st_list.append(st_val)
        lottie_en_list.append(en_val)
        synfig_region(bline, st_val, origin, fr)
        synfig_region(bline, en_val, origin, fr + 1)
        fr += 1
    equalize_length(lottie_st_list, lottie_en_list)
    # Setting final time
    lottie.append({})
    lottie[-1]["t"] = fr

def synfig_region(bline, st_val, origin_p, fr):
    """
    Calculates the points for the region layer in Synfig
    """
    bline_list = bline.get_list_at_frame(fr)
    origin_cur = origin_p.get_value(fr)         

    for bline_point in bline_list:
        tangent1 = bline_point.get_tangent1() * settings.PIX_PER_UNIT
        tangent2 = bline_point.get_tangent2() * settings.PIX_PER_UNIT

        tangent1, tangent2 = convert_tangent_to_lottie(tangent1, tangent2)

        # Adding origin to each vertex
        pos = to_Lottie_axis(bline_point.get_vertex().get_list(), "vector")
        for i in range(len(pos)):
            pos[i] += origin_cur[i]

        # Store values in dictionary
        st_val["i"].append(tangent1.get_list())
        st_val["o"].append(tangent2.get_list())
        st_val["v"].append(pos)



"""
Will store all functions needed to generate the star layer in lottie
"""

import sys
import settings
from properties.value import gen_properties_value
from common.misc import get_angle, change_axis, is_animated
from common.Count import Count
from properties.multiDimensionalKeyframed import gen_properties_multi_dimensional_keyframed
from properties.valueKeyframed import gen_value_Keyframed
sys.path.append("..")


def gen_shapes_star(lottie, layer, idx):
    """
    Generates the dictionary corresponding to shapes/star.json

    Args:
        lottie (dict)               : The lottie generated star layer will be stored in it
        layer  (common.Layer.Layer)  : Synfig format star layer
        idx    (int)                : Stores the index of the star layer

    Returns:
        (None)
    """
    index = Count()
    lottie["ty"] = "sr"     # Type: star
    lottie["pt"] = {}       # Number of points on the star
    lottie["p"] = {}        # Position of star
    lottie["r"] = {}        # Angle / Star's rotation
    lottie["ir"] = {}       # Inner radius
    lottie["or"] = {}       # Outer radius
    lottie["is"] = {}       # Inner roundness of the star
    lottie["os"] = {}       # Outer roundness of the star
    regular_polygon = {"prop" : "false"}

    # Regular polygon
    rp = layer.get_param("regular_polygon").get()
    is_animate = is_animated(rp[0])
    if is_animate == settings.ANIMATED:
        regular_polygon["prop"] = "changing"
        regular_polygon["animated"] = rp[0]
    elif is_animate == settings.SINGLE_WAYPOINT:
        regular_polygon["prop"] = rp[0][0][0].attrib["value"]
    else:
        regular_polygon["prop"] = rp[0].attrib["value"]
    regular_polygon["animate"] = is_animate

    # Points
    points = layer.get_param("points").get()
    is_animate = is_animated(points[0])
    if is_animate == settings.ANIMATED:
        # To uniquely identify the points, attribute type is changed
        points[0].attrib['type'] = 'points'
        gen_value_Keyframed(lottie["pt"], points[0], index.inc())
    else:
        num_points = 3      # default number of points
        if is_animate == settings.NOT_ANIMATED:
            num_points = int(points[0].attrib["value"])
        else:
            num_points = int(points[0][0][0].attrib["value"])
        gen_properties_value(lottie["pt"],
                             num_points,
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)

    # Angle
    angle = layer.get_param("angle").get()
    is_animate = is_animated(angle[0])
    if is_animate == settings.ANIMATED:
        gen_value_Keyframed(lottie["r"], angle[0], index.inc())
    else:
        theta = 0           # default angle for the star
        if is_animate == settings.NOT_ANIMATED:
            theta = get_angle(float(angle[0].attrib["value"]))
        else:
            theta = get_angle(float(angle[0][0][0].attrib["value"]))
        gen_properties_value(lottie["r"],
                             theta,
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)

    # Radius1
    radius1 = layer.get_param("radius1").get()
    is_animate = is_animated(radius1[0])
    if is_animate == settings.ANIMATED:
        gen_value_Keyframed(lottie["or"], radius1[0], index.inc())
    else:
        r_outer = 0             # default value for outer radius
        if is_animate == settings.NOT_ANIMATED:
            r_outer = float(radius1[0].attrib["value"])
        else:
            r_outer = float(radius1[0][0][0].attrib["value"])
        gen_properties_value(lottie["or"],
                             int(settings.PIX_PER_UNIT * r_outer),
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)

    # Radius2
    radius2 = layer.get_param("radius2").get()
    is_animate = is_animated(radius2[0])
    if is_animate == settings.ANIMATED:
        gen_value_Keyframed(lottie["ir"], radius2[0], index.inc())
    else:
        r_inner = 0             # default value for inner radius
        if is_animate == settings.NOT_ANIMATED:
            r_inner = float(radius2[0].attrib["value"])
        else:
            r_inner = float(radius2[0][0][0].attrib["value"])
        gen_properties_value(lottie["ir"],
                             int(settings.PIX_PER_UNIT * r_inner),
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)

    # Origin
    origin = layer.get_param("origin").get()
    is_animate = is_animated(origin[0])
    if is_animate == settings.ANIMATED:
        gen_properties_multi_dimensional_keyframed(lottie["p"],
                                                   origin[0],
                                                   index.inc())
    else:
        x_val, y_val = 0, 0
        if is_animate == settings.NOT_ANIMATED:
            x_val = float(origin[0][0].text) * settings.PIX_PER_UNIT
            y_val = float(origin[0][1].text) * settings.PIX_PER_UNIT
        else:
            x_val = float(origin[0][0][0][0].text) * settings.PIX_PER_UNIT
            y_val = float(origin[0][0][0][1].text) * settings.PIX_PER_UNIT
        gen_properties_value(lottie["p"],
                             change_axis(x_val, y_val),
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)

    # If not animated, then go to if, else
    if regular_polygon["animate"] in {0, 1}:
        if regular_polygon["prop"] == "false":
            lottie["sy"] = 1    # Star Type

            # inner property is only needed if type is star
            gen_properties_value(lottie["is"],
                                 0,
                                 index.inc(),
                                 settings.DEFAULT_ANIMATED,
                                 settings.NO_INFO)
        else:
            lottie["sy"] = 2    # Polygon Type

            # for polygon type, "ir" and "is" must not be present
            del lottie["ir"]

    # If animated, it will always be of type star
    else:
        lottie["sy"] = 1
        gen_properties_value(lottie["is"],
                             0,
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)

    gen_properties_value(lottie["os"],
                         0,
                         index.inc(),
                         settings.DEFAULT_ANIMATED,
                         settings.NO_INFO)
    lottie["ix"] = idx
